# Copyright 2022 Thomas Leberbauer
#           2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from collections.abc import Callable
from gi.repository import GObject, Gio

from quodlibet import _, app, util
from quodlibet.library.song import SongLibrary
from quodlibet.qltk.models import ObjectModelSort, ObjectStore, ObjectModelFilter
from quodlibet.util.collection import Album
from quodlibet.util.i18n import numeric_phrase
from quodlibet.util.library import background_filter


class AlbumListItem(GObject.Object):
    """This model represents an entry for a specific album.

    It will load the album cover and generate the album label on demand.
    """

    def __init__(self, album: Album | None = None):
        super().__init__()
        self._album = album
        self._cover = None
        self._label = None

        self.connect("notify::album", self._album_changed)

    def load_cover(self, size: int, cancelable: Gio.Cancellable | None = None):
        def callback(cover):
            self._cover = cover
            self.notify("cover")

        manager = app.cover_manager
        # Skip this during testing
        if manager:
            manager.get_pixbuf_many_async(
                self._album.songs, size, size, cancelable, callback
            )

    def format_label(self, pattern):
        self._label = pattern % self._album
        self.notify("label")

    def _album_changed(self, model, prop):
        self._label = None

    @GObject.Property
    def album(self):
        return self._album

    @GObject.Property
    def cover(self):
        return self._cover

    @GObject.Property
    def label(self):
        return self._label


class AlbumListCountItem(AlbumListItem):
    """This model represents an entry for a set of albums.

    It will generate a label containing the number of albums on demand.
    """

    def load_cover(self, *args, **kwargs):
        self.notify("cover")

    def format_label(self, pattern=None):
        n = self.__n_albums
        title = util.bold(_("All Albums"))
        number_phrase = numeric_phrase("%d album", "%d albums", n)
        self._label = f"{title}\n{number_phrase}"
        self.notify("label")

    @GObject.Property
    def n_albums(self):
        return self.__n_albums

    @n_albums.setter  # type: ignore
    def n_albums(self, value):
        self.__n_albums = value
        self.format_label()


class AlbumListModel(ObjectStore):
    """This model creates entries for albums from a library.

    The first entry represents the whole set of albums in the library.
    """

    def __init__(self, library: SongLibrary):
        super().__init__()
        self.__library = library

        albums = library.albums
        self.__sigs = [
            albums.connect("added", self._add_albums),
            albums.connect("removed", self._remove_albums),
            albums.connect("changed", self._change_albums),
        ]

        self.append(row=[AlbumListCountItem()])
        self.append_many(AlbumListItem(a) for a in albums.values())

    def destroy(self):
        albums = self.__library.albums
        for sig in self.__sigs:
            albums.disconnect(sig)
        self.__library = None
        self.clear()

    def _add_albums(self, library, added):
        self.append_many(AlbumListItem(a) for a in added)

    def _remove_albums(self, library, removed):
        removed_albums = removed.copy()
        iters_remove = []
        for iter_, item in self.iterrows():
            album = item.album
            if album is not None and album in removed_albums:
                removed_albums.remove(album)
                iters_remove.append(iter_)
                if not removed_albums:
                    break
        for iter_ in iters_remove:
            self.remove(iter_)

    def _change_albums(self, library, changed):
        changed_albums = changed.copy()
        for iter_, item in self.iterrows():
            album = item.album
            if album is not None and album in changed_albums:
                changed_albums.remove(album)
                self.iter_changed(iter_)
                if not changed_albums:
                    break


class AlbumListFilterModel(GObject.Object, Gio.ListModel):
    """This model filters entries in a child model.

    The property "include_item_all" toggles visibility of the first entry of the
    child model.

    The property "filter" is a function which defines visibility for all
    remaining entries of the child model. If "filter" is set to None, all
    entries are visible.
    """

    __item_all: AlbumListItem
    __include_item_all: bool
    __filter: Callable[[AlbumListItem], bool] | None = None

    def __init__(self, child_model=None, include_item_all=True, **kwargs):
        super().__init__(**kwargs)

        self.__include_item_all = include_item_all

        self._model = model = ObjectModelFilter(child_model=child_model)
        self.__item_all = self._get_item(0)
        self._update_n_albums()

        # Tell the tree model that all nodes are visible, otherwise it does not
        # emit the "row-changed" signal.
        for row in model:
            model.ref_node(row.iter)

        model.set_visible_func(self._apply_filter)
        self.__sigs = [
            model.connect("row-changed", self._row_changed),
            model.connect("row-inserted", self._row_inserted),
            model.connect("row-deleted", self._row_deleted),
            model.connect("rows-reordered", self._rows_reordered),
        ]

    def destroy(self):
        model = self._model
        for sig in self.__sigs:
            model.disconnect(sig)
        self._model = None

    @GObject.Property
    def include_item_all(self):
        return self.__include_item_all

    @include_item_all.setter  # type: ignore
    def include_item_all(self, value):
        if self.__include_item_all == value:
            return
        self.__include_item_all = value
        removed, added = (0, 1) if value else (1, 0)
        self.items_changed(0, removed, added)

    @GObject.Property
    def filter(self):
        return self.__filter

    @filter.setter  # type: ignore
    def filter(self, value):
        b = background_filter()
        if b is None and value is None:
            f = None
        elif b is None:

            def f(item):
                return value(item.album)
        else:

            def f(item):
                return b(item.album) and value(item.album)

        if f or self.__filter:
            self.__filter = f
            self._model.refilter()

    def do_get_n_items(self):
        return len(self)

    def do_get_item(self, index):
        return self[index]

    def do_get_item_type(self):
        return AlbumListItem

    def __len__(self):
        n = len(self._model)
        if self.__include_item_all or n < 1:
            return n
        return n - 1

    def __getitem__(self, index):
        if not self.__include_item_all:
            index += 1
        return self._get_item(index)

    def _get_item(self, index: int) -> AlbumListItem | None:
        model = self._model
        iter = model.iter_nth_child(None, index)
        return model.get_value(iter) if iter else None

    def _update_n_albums(self):
        self.__item_all.props.n_albums = len(self._model) - 1

    def _apply_filter(self, model, iter, _):
        filter = self.__filter
        if filter is None:
            return True
        item = model.get_value(iter)
        if item is self.__item_all:
            return True
        return filter(item)

    def _row_changed(self, model, path, iter):
        item = model.get_value(iter)
        item.notify("album")

    def _row_inserted(self, model, path, iter):
        # Ensure that the tree model will emit the "row-changed" signal
        model.ref_node(iter)

        index = path.get_indices()[0]
        if not self.__include_item_all:
            index -= 1
        if index >= 0:
            self.items_changed(index, 0, 1)
            self._update_n_albums()

    def _row_deleted(self, model, path):
        index = path.get_indices()[0]
        if not self.__include_item_all:
            index -= 1
        if index >= 0:
            self.items_changed(index, 1, 0)
            self._update_n_albums()

    def _rows_reordered(self, model, path, iter, new_order):
        n = len(self)
        self.items_changed(0, n, n)


class AlbumListSortModel(ObjectModelSort):
    """This model sorts entries of a child model"""



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, GLib, Pango

from senf import fsnative

from quodlibet import _
from quodlibet import qltk
from quodlibet import util

from quodlibet.qltk import Button, Icons
from quodlibet.formats import AudioFile
from quodlibet.pattern import XMLFromPattern, XMLFromMarkupPattern, Error
from quodlibet.util import connect_obj

try:
    import gi

    gi.require_version("GtkSource", "4")
    from gi.repository import GtkSource
except (ValueError, ImportError):
    TextView = Gtk.TextView
    TextBuffer = Gtk.TextBuffer
else:
    TextView = GtkSource.View

    class TextBuffer(GtkSource.Buffer):  # type: ignore
        def __init__(self, *args):
            super().__init__(*args)
            self.set_highlight_matching_brackets(False)
            self.set_highlight_syntax(False)

        def set_text(self, *args):
            self.begin_not_undoable_action()
            super().set_text(*args)
            self.end_not_undoable_action()


class TextEditBox(Gtk.HBox):
    """A simple text editing area with a default value, a revert button,
    and an apply button. The 'buffer' attribute is the text buffer, the
    'apply' attribute is the apply button.

    FIXME: Button text should changeable (without poking the buttons directly).
    """

    def __init__(self, default=""):
        super().__init__(spacing=6)

        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.add(TextView(buffer=TextBuffer()))
        self.pack_start(sw, True, True, 0)
        self.buffer = sw.get_child().get_buffer()

        box = Gtk.VBox(spacing=6)
        rev = Button(_("_Revert"), Icons.DOCUMENT_REVERT)
        app = Button(_("_Apply"))
        box.pack_start(rev, False, True, 0)
        box.pack_start(app, False, True, 0)
        self.pack_start(box, False, True, 0)
        connect_obj(rev, "clicked", self.buffer.set_text, default)
        self.revert = rev
        self.apply = app

    @property
    def text(self):
        start, end = self.buffer.get_bounds()
        return self.buffer.get_text(start, end, True)

    @text.setter
    def text(self, value):
        self.buffer.set_text(value, -1)


def validate_markup_pattern(text, alternative_markup=True, links=False):
    """Check whether a passed pattern results in a valid pango markup.

    Args:
        text (unicode): the pattern
        alternative_markup (bool): if "[b]" gets mapped to "\\<b\\>"
        links (bool): if link tags are allowed (for Gtk.Label only)

    Raises:
        ValueError: In case the pattern isn't valid
    """

    assert isinstance(text, str)

    f = AudioFile({"~filename": fsnative("dummy")})

    try:
        if alternative_markup:
            pattern = XMLFromMarkupPattern(text)
        else:
            pattern = XMLFromPattern(text)
        text = pattern % f
    except Error as e:
        return ValueError(e)

    try:
        Pango.parse_markup(text, -1, "\u0000")
    except GLib.GError as e:
        if not links:
            raise ValueError(e) from e
        # Gtk.Label supports links on top of pango markup but doesn't
        # provide a way to verify them. We can check if the markup
        # was accepted by seeing if get_text() returns something.
        l = Gtk.Label()
        # add a character in case text is empty.
        # this might print a warning to stderr.. no idea how to prevent that..
        l.set_markup(text + " ")
        if not l.get_text():
            raise ValueError(e) from e


class PatternEditBox(TextEditBox):
    """A TextEditBox that stops the apply button's clicked signal if
    the pattern is invalid. You need to use connect_after to connect to
    it, to get this feature."""

    def __init__(self, default="", alternative_markup=True, links=False):
        super().__init__(default)
        self._alternative_markup = alternative_markup
        self._links = links
        self.apply.connect("clicked", self.__check_markup)

    def __check_markup(self, apply):
        try:
            validate_markup_pattern(self.text, self._alternative_markup, self._links)
        except ValueError as e:
            qltk.ErrorMessage(
                self,
                _("Invalid pattern"),
                _(
                    "The pattern you entered was invalid. Make sure you enter "
                    "&lt; and &gt; as \\&lt; and \\&gt; and that your tags are "
                    "balanced.\n\n%s"
                )
                % util.escape(str(e)),
            ).run()
            apply.stop_emission("clicked")
        return False


class TextEdit(qltk.UniqueWindow):
    """A window with a text editing box in it."""

    Box = TextEditBox

    def __init__(self, parent, default="", **kwargs):
        if self.is_not_unique():
            return
        super().__init__()
        self.set_title(_("Edit Display"))
        self.set_transient_for(qltk.get_top_parent(parent))
        self.set_border_width(12)
        self.set_default_size(420, 190)

        vbox = Gtk.VBox(spacing=12)
        close = Button(_("_Close"), Icons.WINDOW_CLOSE)
        close.connect("clicked", lambda *x: self.destroy())
        b = Gtk.HButtonBox()
        b.set_layout(Gtk.ButtonBoxStyle.END)
        b.pack_start(close, True, True, 0)

        self.box = box = self.Box(default, **kwargs)
        vbox.pack_start(box, True, True, 0)
        self.use_header_bar()
        if not self.has_close_button():
            vbox.pack_start(b, False, True, 0)

        self.add(vbox)
        self.apply = box.apply
        self.revert = box.revert

        close.grab_focus()
        self.get_child().show_all()

    @property
    def text(self):
        return self.box.text

    @text.setter
    def text(self, value):
        self.box.text = value


class PatternEdit(TextEdit):
    """A window with a pattern editing box in it."""

    Box = PatternEditBox



# Copyright 2013 Christoph Reiter
#           2015 Nick Boultbee
#           2017 Fredrik Strupe
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import config
from quodlibet import util
from quodlibet import qltk
from quodlibet import _
from quodlibet.qltk.views import BaseView
from quodlibet.qltk.tagscombobox import TagsComboBoxEntry
from quodlibet.qltk.x import SymbolicIconImage, MenuItem, Button
from quodlibet.qltk import Icons
from quodlibet.qltk.menubutton import MenuButton
from quodlibet.qltk.ccb import ConfigCheckButton
from quodlibet.util import connect_obj
from .util import get_headers, save_headers


@util.enum
class ColumnMode(int):
    SMALL = 0
    WIDE = 1
    COLUMNAR = 2


class ColumnModeSelection(Gtk.VBox):
    def __init__(self, browser):
        super().__init__(spacing=6)
        self.browser = browser
        self.buttons = []

        group = None
        mode_label = {
            ColumnMode.SMALL: _("Small"),
            ColumnMode.WIDE: _("Wide"),
            ColumnMode.COLUMNAR: _("Columnar"),
        }
        for mode in ColumnMode.values:
            lbl = mode_label[ColumnMode.value_of(mode)]
            group = Gtk.RadioButton(group=group, label=lbl)
            if mode == config.getint("browsers", "pane_mode", ColumnMode.SMALL):
                group.set_active(True)
            self.pack_start(group, False, True, 0)
            self.buttons.append(group)

        # Connect to signal after the correct radio button has been
        # selected
        for button in self.buttons:
            button.connect("toggled", self.toggled)

    def toggled(self, button):
        selected_mode = ColumnMode.SMALL
        if self.buttons[1].get_active():
            selected_mode = ColumnMode.WIDE
        if self.buttons[2].get_active():
            selected_mode = ColumnMode.COLUMNAR
        config.set("browsers", "pane_mode", int(selected_mode))
        self.browser.set_all_column_mode(selected_mode)


class PatternEditor(Gtk.VBox):
    PRESETS = [
        ["genre", "~people", "album"],
        ["~people", "album"],
    ]
    COMPLETION = ["genre", "grouping", "~people", "artist", "album", "~year", "~rating"]

    _COMPLEX_PATTERN_EXAMPLE = "<~year|[b]<~year>[/b]|[i]unknown year[/i]>"

    def __init__(self):
        super().__init__(spacing=6)

        self.__headers = headers = {}
        buttons = []

        group = None
        for tags in self.PRESETS:
            tied = "~" + "~".join(tags)
            group = Gtk.RadioButton(
                group=group, label="_" + util.tag(tied), use_underline=True
            )
            headers[group] = tags
            buttons.append(group)

        group = Gtk.RadioButton(group=group, label=_("_Custom"), use_underline=True)
        self.__custom = group
        headers[group] = []
        buttons.append(group)

        button_box = Gtk.HBox(spacing=6)
        self.__model = model = Gtk.ListStore(str)

        radio_box = Gtk.VBox(spacing=6)
        for button in buttons:
            radio_box.pack_start(button, False, True, 0)
            button.connect("toggled", self.__toggled, button_box, model)

        self.pack_start(radio_box, False, True, 0)

        example = util.monospace(self._COMPLEX_PATTERN_EXAMPLE)
        tooltip = _("Tag pattern with optional markup e.g. %(short)s or\n%(long)s") % {
            "short": "<tt>composer</tt>",
            "long": example,
        }

        cb = TagsComboBoxEntry(self.COMPLETION, tooltip_markup=tooltip)

        view = BaseView(model=model)
        view.set_reorderable(True)
        view.set_headers_visible(False)

        ctrl_box = Gtk.VBox(spacing=6)

        add = Button(_("_Add"), Icons.LIST_ADD)
        ctrl_box.pack_start(add, False, True, 0)
        add.connect("clicked", self.__add, model, cb)

        remove = Button(_("_Remove"), Icons.LIST_REMOVE)
        ctrl_box.pack_start(remove, False, True, 0)
        remove.connect("clicked", self.__remove, view)

        selection = view.get_selection()
        selection.connect("changed", self.__selection_changed, remove)
        selection.emit("changed")

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.add(view)

        edit_box = Gtk.VBox(spacing=6)
        edit_box.pack_start(cb, False, True, 0)
        edit_box.pack_start(sw, True, True, 0)

        button_box.pack_start(edit_box, True, True, 0)
        button_box.pack_start(ctrl_box, False, True, 0)
        self.pack_start(button_box, True, True, 0)

        render = Gtk.CellRendererText()
        render.set_property("editable", True)

        def edited_cb(render, path, text, model):
            model[path][0] = text

        render.connect("edited", edited_cb, model)

        column = Gtk.TreeViewColumn(None, render, text=0)
        view.append_column(column)

    @property
    def headers(self):
        for button in self.__headers.keys():
            if button.get_active():
                if button == self.__custom:
                    model_headers = [row[0] for row in self.__model]
                    self.__headers[self.__custom] = model_headers
                return self.__headers[button]
        return None

    @headers.setter
    def headers(self, new_headers):
        for button, headers in self.__headers.items():
            if headers == new_headers:
                button.set_active(True)
                button.emit("toggled")
                break
        else:
            self.__headers[self.__custom] = new_headers
            self.__custom.set_active(True)

    def __selection_changed(self, selection, remove):
        remove.set_sensitive(bool(selection.get_selected()[1]))

    def __add(self, button, model, cb):
        if cb.tag:
            model.append(row=[cb.tag])

    def __remove(self, button, view):
        view.remove_selection()

    def __toggled(self, button, edit_widget, model):
        tags = self.__headers[button]

        if tags:
            model.clear()
            for h in tags:
                model.append(row=[h])

        edit_widget.set_sensitive(button.get_active() and button is self.__custom)


class PreferencesButton(Gtk.HBox):
    def __init__(self, browser):
        super().__init__()

        self._menu = menu = Gtk.Menu()

        pref_item = MenuItem(_("_Preferences"), Icons.PREFERENCES_SYSTEM)

        def preferences_cb(menu_item):
            window = Preferences(browser)
            window.show()

        pref_item.connect("activate", preferences_cb)
        menu.append(pref_item)

        menu.show_all()

        button = MenuButton(
            SymbolicIconImage(Icons.EMBLEM_SYSTEM, Gtk.IconSize.MENU), arrow=True
        )
        button.set_menu(menu)
        button.show()
        self.pack_start(button, True, True, 0)


class Preferences(qltk.UniqueWindow):
    def __init__(self, browser):
        if self.is_not_unique():
            return
        super().__init__()

        self.set_transient_for(qltk.get_top_parent(browser))
        self.set_default_size(350, 300)
        self.set_border_width(12)

        self.set_title(_("Paned Browser Preferences"))

        vbox = Gtk.VBox(spacing=12)

        column_modes = ColumnModeSelection(browser)
        column_mode_frame = qltk.Frame(_("Column layout"), child=column_modes)

        editor = PatternEditor()
        editor.headers = get_headers()
        editor_frame = qltk.Frame(_("Column content"), child=editor)

        equal_width = ConfigCheckButton(
            _("Equal pane width"), "browsers", "equal_pane_width", populate=True
        )

        apply_ = Button(_("_Apply"))
        connect_obj(
            apply_, "clicked", self.__apply, editor, browser, False, equal_width
        )

        cancel = Button(_("_Cancel"))
        cancel.connect("clicked", lambda x: self.destroy())

        box = Gtk.HButtonBox()
        box.set_spacing(6)
        box.set_layout(Gtk.ButtonBoxStyle.EDGE)
        box.pack_start(equal_width, True, True, 0)
        box.pack_start(apply_, False, False, 0)
        self.use_header_bar()
        if not self.has_close_button():
            box.pack_start(cancel, True, True, 0)

        vbox.pack_start(column_mode_frame, False, False, 0)
        vbox.pack_start(editor_frame, True, True, 0)
        vbox.pack_start(box, False, True, 0)

        self.add(vbox)

        cancel.grab_focus()
        self.get_child().show_all()

    def __apply(self, editor, browser, close, equal_width):
        if editor.headers != get_headers():
            save_headers(editor.headers)
            browser.set_all_panes()

        if equal_width.get_active():
            browser.make_pane_widths_equal()

        if close:
            self.destroy()



# Copyright 2014 Christoph Reiter
#           2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


def enum(cls):
    """Class decorator for enum types::

        @enum
        class SomeEnum(int):
            FOO = 0
            BAR = 1

    Result is an int subclass and all attributes are instances of it.
    Each subclass has a property `values` referring to *all* known instances.
    """

    type_ = cls.__bases__[0]

    d = dict(cls.__dict__)
    new_type = type(cls.__name__, (type_,), d)
    new_type.__module__ = cls.__module__

    map_ = {}
    for key, value in d.items():
        if key.upper() == key:
            value_instance = new_type(value)
            setattr(new_type, key, value_instance)
            map_[value] = key
    new_type.values = set(map_.keys())

    def value_of(cls, s, default=None):
        for v in cls.values:
            if v == s:
                return v
        if default is not None:
            return default
        raise ValueError(f"Can't find {s} (try {cls.values})")

    new_type.value_of = classmethod(value_of)

    def repr_(self):
        name = type(self).__name__
        try:
            return f"{name}.{map_[self]}"
        except KeyError:
            return f"{name}({self})"

    new_type.__repr__ = repr_

    return new_type



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Extension points for Quod Libet and Ex Falso

This (currently) contains code plugins (formerly kept in /plugins)
arranged by plugin type (songsmenu, events, etc).

Base classes themselves are in /quodlibet/plugins
"""



# Copyright 2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from mutagen.aiff import AIFF

from ._id3 import ID3File

extensions = [".aif", ".aiff", ".aifc"]


class AIFFFile(ID3File):
    format = "AIFF"

    mimes = ["audio/x-aiff", "audio/aiff"]
    Kind = AIFF

    def _parse_info(self, info):
        self["~#length"] = info.length
        self["~#bitrate"] = int(info.bitrate / 1000)
        self["~#channels"] = info.channels
        self["~#samplerate"] = info.sample_rate


loader = AIFFFile
types = [AIFFFile]



# Copyright 2018 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


class SessionError(Exception):
    pass


class SessionClient:
    def open(self, app):
        """Raises SessionError"""

    def close(self):
        """Doesn't raise, can be called multiple times"""



# Copyright 2018 David Morris
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk
from quodlibet import _
from quodlibet import app
from quodlibet import qltk
from quodlibet.plugins.events import EventPlugin
from quodlibet.plugins import PluginConfigMixin
from quodlibet.qltk import Icons
from quodlibet.qltk.ccb import ConfigCheckButton


class StopOnPause(EventPlugin, PluginConfigMixin):
    PLUGIN_ID = "StopOnPause"
    PLUGIN_NAME = _("Stop on Pause")
    PLUGIN_ICON = Icons.MEDIA_PLAYBACK_PAUSE
    PLUGIN_DESC = _(
        "This plugin changes Pause to Stop/Play/Seek."
        "\n\n"
        "Useful if Quod Libet can't switch audio devices "
        "when a new one (Bluetooth speaker, USB DAC, etc.) "
        "is added to the system."
        "\n\n"
        "Ensures position in the current source is maintained "
        "so long as 'Only Seekable Sources' is off. "
        "Otherwise, playback is left stopped and must be restarted "
        "from the beginning."
    )

    _CFG_SEEKABLE_ONLY = "seekable_only"

    __enabled = False
    __restarting = False

    def enabled(self):
        self.__enabled = True

    def disabled(self):
        self.__enabled = False

    def plugin_on_paused(self):
        if self.__enabled:
            if not self.__restarting:
                only_seekable = self.config_get_bool(self._CFG_SEEKABLE_ONLY)

                if only_seekable and app.player.seekable:
                    self.__restarting = True
                    position = app.player.get_position()
                    app.player.stop()
                    app.player.seek(position)
                    self.__restarting = False
                elif not only_seekable:
                    self.__restarting = True

                    # Check if the stream is seekable before stopping!
                    is_seekable = app.player.seekable
                    position = app.player.get_position()

                    app.player.stop()

                    if is_seekable:
                        app.player.seek(position)

                    self.__restarting = False

    @classmethod
    def PluginPreferences(cls, window):
        # Plugin Options
        toggles = [
            (cls._CFG_SEEKABLE_ONLY, _("Only Seekable Sources")),
        ]

        vb = Gtk.VBox(spacing=10)
        vb.set_border_width(0)
        vb2 = Gtk.VBox(spacing=6)
        for key, label in toggles:
            ccb = ConfigCheckButton(label, "plugins", cls._config_key(key))
            ccb.set_active(cls.config_get_bool(key))
            vb2.pack_start(ccb, True, True, 0)

        frame = qltk.Frame(label=_("Plugin Options"), child=vb2)
        vb.pack_start(frame, False, True, 0)

        vb.show_all()
        return vb



# Copyright 2006 Joe Wreschnig
#           2011-2024 Nick Boultbee
#           2013,2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import os
import shutil
from typing import TypeVar, Optional, Generic
from collections.abc import (
    Collection,
    Sequence,
    Iterable,
    Iterator,
    MutableMapping,
    Generator,
)

from gi.repository import GObject

import quodlibet
from quodlibet import util
from quodlibet.formats import load_audio_files, dump_audio_files, SerializationError
from quodlibet.formats._audio import HasKey
from quodlibet.util.atomic import atomic_save
from quodlibet.util.collections import DictMixin
from quodlibet.util.dprint import print_d, print_w
from quodlibet.util.path import mkdir, is_hidden
from senf import fsnative, path2fsn

K = TypeVar("K", covariant=True)
V = TypeVar("V", bound=HasKey)


class Library(GObject.GObject, DictMixin, Generic[K, V]):
    """A Library contains useful objects.

    The only required method these objects support is a .key
    attribute, but specific types of libraries may require more
    advanced interfaces.

    Every method which takes a sequence of items expects items to
    implement __iter__, __len__ and __contains__.

    Likewise the signals emit sequences which implement
    __iter__, __len__ and __contains__ e.g. set(), list() or tuple().

    WARNING: The library implements the dict interface with the exception
    that iterating over it yields values and not keys.
    """

    __gsignals__ = {
        "changed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "removed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "added": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    librarian: Optional["quodlibet.library.librarians.Librarian"] = None
    """A librarian, if defined will be used for collaborating with other libraries"""

    dirty = False

    def __init__(self, name: str | None = None):
        super().__init__()
        self._contents: MutableMapping[K, V] = {}
        self._name = name
        if self.librarian is not None and name is not None:
            self.librarian.register(self, name)

    def __str__(self):
        return f"<{type(self).__name__} @ {hex(id(self))}>"

    def destroy(self):
        if self.librarian is not None and self._name is not None:
            self.librarian._unregister(self, self._name)

    def changed(self, items: Collection[V]):
        """Alert other users that these items have changed.

        This causes a 'changed' signal. If a librarian is available
        this function will call its changed method instead, and all
        libraries that librarian manages may fire a 'changed' signal.

        The item list may be filtered to those items actually in the
        library. If a librarian is available, it will handle the
        filtering instead. That means if this method is delegated to
        the librarian, this library's changed signal may not fire, but
        another's might.
        """

        if not items:
            return
        if self.librarian and self in self.librarian.libraries.values():
            print_d(f"Changing {len(items)} items via librarian.", self._name)
            self.librarian.changed(items)
        else:
            items = {item for item in items if item in self}
            if not items:
                return
            print_d(f"Changing {len(items)} items directly.", self._name)
            self._changed(items)

    def _changed(self, items: Collection[V]):
        """Called by the changed method and Librarians."""

        if not items:
            return
        print_d(
            f"Emitting changed for {len(items)} item(s) "
            f"(e.g. {list(items)[0].key!r}...)",
            self._name,
        )
        self.dirty = True
        self.emit("changed", items)

    def __iter__(self) -> Iterator[V]:
        """Iterate over the items in the library."""
        return iter(self._contents.values())

    def iteritems(self) -> Iterator[tuple[K, V]]:
        return iter(self._contents.items())

    def iterkeys(self) -> Iterator[K]:
        return iter(self._contents.keys())

    def itervalues(self) -> Iterator[V]:
        return iter(self._contents.values())

    def __len__(self) -> int:
        """The number of items in the library."""
        return len(self._contents)

    def __getitem__(self, key) -> V:
        """Find a item given its key."""
        return self._contents[key]

    def __contains__(self, item):
        """Check if a key or item is in the library."""
        try:
            return item in self._contents or item.key in self._contents
        except AttributeError:
            return False

    def get_content(self) -> Sequence[V]:
        """All items including hidden ones for saving the library
        (see FileLibrary with masked items)
        """
        return list(self.values())

    def keys(self) -> Iterable[K]:
        return self._contents.keys()

    def values(self) -> Iterable[V]:
        return self._contents.values()

    def _load_item(self, item: V) -> None:
        """Load (add) an item into this library"""
        # Subclasses should override this if they want to check
        # item validity; see `FileLibrary`.
        print_d(f"Loading {item.key!r}", self._name)
        self.dirty = True
        self._contents[item.key] = item

    def _load_init(self, items: Iterable[V]) -> None:
        """Load many items into the library (on start)"""
        # Subclasses should override this if they want to check
        # item validity; see `FileLibrary`.
        content = self._contents
        for item in items:
            content[item.key] = item

    def add(self, items: Iterable[V]) -> set[V]:
        """Add items. This causes an 'added' signal.

        Return the sequence of items actually added, filtering out items
        already in the library.
        """

        items = {item for item in items if item is not None and item not in self}
        if not items:
            return items
        if len(items) == 1:
            print_d(f"Adding {next(iter(items))}", self._name)
        else:
            print_d(f"Adding {len(items)} items.", self._name)
        for item in items:
            self._contents[item.key] = item

        self.dirty = True
        self.emit("added", items)
        return items

    def remove(self, items: Iterable[V]) -> set[V]:
        """Remove items. This causes a 'removed' signal.

        Return the sequence of items actually removed.
        """

        removals = {item for item in items if item in self}
        if not removals:
            return removals

        print_d(f"Removing {len(removals)} item(s).", self._name)
        for item in removals:
            del self._contents[item.key]

        self.dirty = True
        self.emit("removed", removals)
        return removals


def _load_items(filename) -> Iterable[V]:
    """Load items from disk.

    In case of an error returns default or an empty list.
    """

    try:
        with open(filename, "rb") as fp:
            data = fp.read()
    except OSError:
        print_w(f"Couldn't load library file from: {filename!r}")
        return []

    try:
        items = load_audio_files(data)
    except SerializationError:
        # there are too many ways this could fail
        util.print_exc()

        # move the broken file out of the way
        try:
            shutil.copy(filename, filename + ".not-valid")
        except OSError:
            util.print_exc()

        return []

    return items


class PicklingMixin:
    """A mixin to provide persistence of a library by pickling to disk"""

    filename = None

    def load(self, filename):
        """Load a library from a file, containing a picked list.

        Loading does not cause added, changed, or removed signals.
        """

        self.filename = filename
        print_d(f"Loading contents of {filename!r}.", self)

        items = _load_items(filename)

        # this loads all items without checking their validity, but makes
        # sure that non-mounted items are masked
        self._load_init(items)

        print_d(f"Done loading contents of {filename!r}", self._name)

    def save(self, filename=None):
        """Save the library to the given filename, or the default if `None`"""

        if filename is None:
            filename = self.filename

        print_d(f"Saving contents to {filename!r}", self._name)

        try:
            dirname = os.path.dirname(filename)
            mkdir(dirname)
            with atomic_save(filename, "wb") as fileobj:
                fileobj.write(dump_audio_files(self.get_content()))
        except SerializationError:
            # Can happen when we try to pickle while the library is being
            # modified, like in the periodic 15min save.
            # Ignore, as it should try again later or on program exit.
            util.print_exc()
        except OSError:
            print_w(f"Couldn't save library to path {filename!r}")
        else:
            self.dirty = False


def iter_paths(
    root: fsnative, exclude: Iterable[fsnative] | None = None, skip_hidden: bool = True
) -> Generator[fsnative, None, None]:
    """Yields paths contained in root (symlinks dereferenced)

    Any path starting with any of the path parts included in exclude
    are ignored (before and after dereferencing symlinks)

    Directory symlinks are not followed (except root itself)

    Args:
        root: start here
        exclude: ignore any of these
        skip_hidden: Ignore files which are hidden or where any
            of the parent directories are hidden.
    Yields:
        fsnative: absolute dereferenced paths
    """

    assert isinstance(root, fsnative)
    exclude = exclude or []
    assert all(isinstance(p, fsnative) for p in exclude)
    assert os.path.abspath(root)

    def skip(path):
        if skip_hidden and is_hidden(path):
            return True
        # FIXME: normalize paths..
        return any(path.startswith(p) for p in exclude)

    if skip_hidden and is_hidden(root):
        return

    for path, dnames, fnames in os.walk(root):
        if skip_hidden:
            dnames[:] = [
                d for d in dnames if not is_hidden(path2fsn(os.path.join(path, d)))
            ]
        for filename in fnames:
            full_filename = path2fsn(os.path.join(path, filename))
            if skip(full_filename):
                continue
            full_filename = path2fsn(os.path.realpath(full_filename))
            if skip(full_filename):
                continue
            yield full_filename  # type: ignore



# Copyright 2004-2005 Joe Wreschnig, Michael Urman
#           2009-2014 Christoph Reiter
#                2021 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
import base64


import mutagen
from mutagen.flac import Picture, error as FLACError  # noqa
from mutagen.id3 import ID3

from quodlibet import config
from quodlibet import const
from quodlibet.util.path import get_temp_cover_file

from ._audio import AudioFile, translate_errors, AudioFileError
from ._image import EmbeddedImage, APICType


# Migrate old layout
sys.modules["formats.flac"] = sys.modules[__name__]
sys.modules["formats.oggvorbis"] = sys.modules[__name__]


class MutagenVCFile(AudioFile):
    format = "Unknown Mutagen + vorbiscomment"
    MutagenType: type[mutagen.FileType] | None = None

    supports_rating_and_play_count_in_file = True

    can_change_images = True

    def __init__(self, filename, audio=None):
        # If we're done a type probe, use the results of that to avoid
        # reopening the file.
        if audio is None:
            with translate_errors():
                audio = self.MutagenType(filename)
        self["~#length"] = audio.info.length
        try:
            self["~#bitrate"] = int(audio.info.bitrate / 1000)
        except AttributeError:
            pass
        try:
            self["~#channels"] = audio.info.channels
        except AttributeError:
            pass
        try:
            self["~#samplerate"] = audio.info.sample_rate
        except AttributeError:
            pass
        if audio.tags and audio.tags.vendor:
            self["~encoding"] = audio.tags.vendor
        # mutagen keys are lower cased
        for key, value in (audio.tags or {}).items():
            self[key] = "\n".join(value)
        self.__post_read()
        self.sanitize(filename)

    def __post_read_total(self, main, fallback, single):
        one = None
        total = None

        if single in self:
            parts = self[single].split("/", 1)
            if parts[0]:
                one = parts[0]
            if len(parts) > 1:
                total = parts[1]
            del self[single]

        if main in self:
            total = self[main]
            del self[main]
        else:
            if fallback in self:
                total = self[fallback]
                del self[fallback]

        final = None
        if one is not None:
            final = one
        if total is not None:
            final = f"{final or ''}/{total}"

        if final is not None:
            self[single] = final

    def __post_read(self):
        email = config.get("editing", "save_email").strip()
        maps = {"rating": float, "playcount": int}
        for keyed_key, func in maps.items():
            emails = [s.lower() for s in ["", f":{const.EMAIL}", f":{email}"]]
            for subkey in emails:
                key = keyed_key + subkey
                if key in self:
                    try:
                        self[f"~#{keyed_key}"] = func(self[key])
                    except ValueError:
                        pass
                    del self[key]

        if "metadata_block_picture" in self:
            self.has_images = True
            del self["metadata_block_picture"]

        if "coverart" in self:
            self.has_images = True
            del self["coverart"]

        if "coverartmime" in self:
            del self["coverartmime"]

        self.__post_read_total("tracktotal", "totaltracks", "tracknumber")
        self.__post_read_total("disctotal", "totaldiscs", "discnumber")

    def get_images(self):
        try:
            audio = self.MutagenType(self["~filename"])
        except Exception:
            return []

        # metadata_block_picture
        images = []
        for data in audio.get("metadata_block_picture", []):
            try:
                cover = Picture(base64.b64decode(data))
            except (TypeError, FLACError):
                continue

            f = get_temp_cover_file(cover.data)
            images.append(
                EmbeddedImage(
                    f, cover.mime, cover.width, cover.height, cover.depth, cover.type
                )
            )

        # coverart + coverartmime
        cover = audio.get("coverart")
        try:
            cover = cover and base64.b64decode(cover[0])
        except TypeError:
            cover = None

        if cover:
            mime = audio.get("coverartmime")
            mime = (mime and mime[0]) or "image/"
            f = get_temp_cover_file(cover, mime)
            images.append(EmbeddedImage(f, mime))

        images.sort(key=lambda c: c.sort_key)

        return images

    def get_primary_image(self):
        """Returns the primary embedded image"""

        try:
            audio = self.MutagenType(self["~filename"])
        except Exception:
            return None

        pictures = []
        for data in audio.get("metadata_block_picture", []):
            try:
                pictures.append(Picture(base64.b64decode(data)))
            except (TypeError, FLACError, ValueError):
                pass

        cover = None
        for pic in pictures:
            if pic.type == APICType.COVER_FRONT:
                cover = pic
                break
            cover = cover or pic

        if cover:
            f = get_temp_cover_file(cover.data, cover.mime)
            return EmbeddedImage(
                f, cover.mime, cover.width, cover.height, cover.depth, cover.type
            )

        cover = audio.get("coverart")
        try:
            cover = cover and base64.b64decode(cover[0])
        except (TypeError, ValueError):
            cover = None

        if not cover:
            self.has_images = False
            return None

        mime = audio.get("coverartmime")
        mime = (mime and mime[0]) or "image/"
        f = get_temp_cover_file(cover, mime)
        return EmbeddedImage(f, mime)

    def clear_images(self):
        """Delete all embedded images"""

        with translate_errors():
            audio = self.MutagenType(self["~filename"])
            audio.pop("metadata_block_picture", None)
            audio.pop("coverart", None)
            audio.pop("coverartmime", None)
            audio.save()

        self.has_images = False

    def set_image(self, image):
        """Replaces all embedded images by the passed image"""

        with translate_errors():
            audio = self.MutagenType(self["~filename"])

        try:
            data = image.read()
        except OSError as e:
            raise AudioFileError(e) from e

        pic = Picture()
        pic.data = data
        pic.type = APICType.COVER_FRONT
        pic.mime = image.mime_type
        pic.width = image.width
        pic.height = image.height
        pic.depth = image.color_depth

        audio.pop("coverart", None)
        audio.pop("coverartmime", None)
        audio["metadata_block_picture"] = base64.b64encode(pic.write()).decode("ascii")

        with translate_errors():
            audio.save()

        self.has_images = True

    def can_change(self, k=None):
        if k is None:
            return super().can_change(None)
        l = k.lower()
        return (
            super().can_change(k)
            and l
            not in [
                "rating",
                "playcount",
                "metadata_block_picture",
                "coverart",
                "coverartmime",
            ]
            and not l.startswith("rating:")
            and not l.startswith("playcount:")
        )

    def __prep_write(self, comments):
        email = config.get("editing", "save_email").strip()
        for key in comments.keys():
            if key.startswith(("rating:", "playcount:")):
                if key.split(":", 1)[1] in [const.EMAIL, email]:
                    del comments[key]
            elif key not in ["metadata_block_picture", "coverart", "coverartmime"]:
                del comments[key]

        if config.getboolean("editing", "save_to_songs"):
            email = email or const.EMAIL
            if self.has_rating:
                comments[f"rating:{email}"] = str(self("~#rating"))
            playcount = self.get("~#playcount", 0)
            if playcount != 0:
                comments[f"playcount:{email}"] = str(playcount)

    def __prep_write_total(self, comments, main, fallback, single):
        lower = self.as_lowercased()

        for k in [main, fallback, single]:
            if k in comments:
                del comments[k]

        if single in lower:
            parts = lower[single].split("/", 1)

            if parts[0]:
                comments[single] = [parts[0]]

            if len(parts) > 1:
                comments[main] = [parts[1]]

        if main in lower:
            comments[main] = lower.list(main)

        if fallback in lower:
            if main in comments:
                comments[fallback] = lower.list(fallback)
            else:
                comments[main] = lower.list(fallback)

    def has_rating_and_playcount_in_file(self, email):
        with translate_errors():
            audio = self.MutagenType(self["~filename"])
        tags = audio.tags
        if tags is None:
            return False
        return f"rating:{email}" in tags and f"playcount:{email}" in tags

    def write(self):
        with translate_errors():
            audio = self.MutagenType(self["~filename"])
        if audio.tags is None:
            audio.add_tags()

        self.__prep_write(audio.tags)

        lower = self.as_lowercased()
        for key in lower.realkeys():
            audio.tags[key] = lower.list(key)

        self.__prep_write_total(audio.tags, "tracktotal", "totaltracks", "tracknumber")
        self.__prep_write_total(audio.tags, "disctotal", "totaldiscs", "discnumber")

        with translate_errors():
            audio.save()
        self.sanitize()


extensions = []
ogg_formats: list[type[mutagen.FileType]] = []

from mutagen.oggvorbis import OggVorbis

extensions.append(".ogg")
extensions.append(".oga")
ogg_formats.append(OggVorbis)

from mutagen.flac import FLAC, FLACNoHeaderError

extensions.append(".flac")
ogg_formats.append(FLAC)

from mutagen.oggflac import OggFLAC

extensions.append(".oggflac")
ogg_formats.append(OggFLAC)

from mutagen.oggspeex import OggSpeex

extensions.append(".spx")
ogg_formats.append(OggSpeex)

from mutagen.oggtheora import OggTheora

extensions.append(".ogv")
ogg_formats.append(OggTheora)

from mutagen.oggopus import OggOpus

extensions.append(".opus")
ogg_formats.append(OggOpus)


class OggFile(MutagenVCFile):
    format = "Ogg Vorbis"
    mimes = ["audio/vorbis", "audio/ogg; codecs=vorbis"]
    MutagenType = OggVorbis


class OggFLACFile(MutagenVCFile):
    format = "Ogg FLAC"
    mimes = ["audio/x-oggflac", "audio/ogg; codecs=flac"]
    MutagenType = OggFLAC


class OggSpeexFile(MutagenVCFile):
    format = "Ogg Speex"
    mimes = ["audio/x-speex", "audio/ogg; codecs=speex"]
    MutagenType = OggSpeex


class OggTheoraFile(MutagenVCFile):
    format = "Ogg Theora"
    mimes = ["video/x-theora", "video/ogg; codecs=theora"]
    MutagenType = OggTheora


class OggOpusFile(MutagenVCFile):
    format = "Ogg Opus"
    mimes = ["audio/ogg; codecs=opus"]
    MutagenType = OggOpus

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self["~#samplerate"] = 48000


class FLACFile(MutagenVCFile):
    format = "FLAC"
    mimes = ["audio/x-flac", "application/x-flac"]
    MutagenType = FLAC

    def __init__(self, filename, audio=None):
        if audio is None:
            with translate_errors():
                audio = FLAC(filename)
        super().__init__(filename, audio)
        if audio.pictures:
            self.has_images = True
        self["~#bitdepth"] = audio.info.bits_per_sample

    def get_images(self):
        images = super().get_images()

        try:
            tag = FLAC(self["~filename"])
        except Exception:
            return images

        for cover in tag.pictures:
            fileobj = get_temp_cover_file(cover.data, cover.mime)
            images.append(
                EmbeddedImage(
                    fileobj,
                    cover.mime,
                    cover.width,
                    cover.height,
                    cover.depth,
                    cover.type,
                )
            )

        images.sort(key=lambda c: c.sort_key)

        return images

    def get_primary_image(self):
        """Returns the primary embedded image"""

        try:
            tag = FLAC(self["~filename"])
        except Exception:
            return None

        covers = tag.pictures
        if not covers:
            return super().get_primary_image()

        covers.sort(key=lambda c: APICType.sort_key(c.type))
        cover = covers[0]

        fileobj = get_temp_cover_file(cover.data, cover.mime)
        return EmbeddedImage(
            fileobj, cover.mime, cover.width, cover.height, cover.depth, cover.type
        )

    def clear_images(self):
        """Delete all embedded images"""

        with translate_errors():
            tag = FLAC(self["~filename"])
            tag.clear_pictures()
            tag.save()

        # clear vcomment tags
        super().clear_images()

        self.has_images = False

    def set_image(self, image):
        """Replaces all embedded images by the passed image"""

        with translate_errors():
            tag = FLAC(self["~filename"])

        try:
            data = image.read()
        except OSError as e:
            raise AudioFileError(e) from e

        pic = Picture()
        pic.data = data
        pic.type = APICType.COVER_FRONT
        pic.mime = image.mime_type
        pic.width = image.width
        pic.height = image.height
        pic.depth = image.color_depth

        tag.add_picture(pic)

        with translate_errors():
            tag.save()

        # clear vcomment tags
        super().clear_images()

        self.has_images = True

    def write(self):
        if ID3 is not None:
            with translate_errors():
                ID3().delete(filename=self["~filename"])
        super().write()


types = []
for var in list(globals().values()):
    if getattr(var, "MutagenType", None):
        types.append(var)


def loader(filename):
    """
    Returns:
        AudioFile
    Raises:
        AudioFileError
    """

    with translate_errors():
        audio = mutagen.File(filename, options=ogg_formats)
        if audio is None and FLAC is not None:
            # FLAC with ID3
            try:
                audio = FLAC(filename)
            except FLACNoHeaderError:
                pass
        if audio is None:
            raise AudioFileError("file type could not be determined")
        audio_cls = type(audio)
        for klass in globals().values():
            if audio_cls is getattr(klass, "MutagenType", None):
                return klass(filename, audio)
        raise AudioFileError("file type could not be determined")



# Copyright 2011,2014 Christoph Reiter <reiter.christoph@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

from gi.repository import Gio
from gi.repository import GLib

from quodlibet import _
from quodlibet import app
from quodlibet.qltk import Icons
from quodlibet.plugins.events import EventPlugin


class ScreensaverPause(EventPlugin):
    PLUGIN_ID = "screensaver_pause"
    PLUGIN_NAME = _("Screensaver Pause")
    PLUGIN_DESC = _("Pauses playback while the GNOME screensaver is active.")
    PLUGIN_ICON = Icons.MEDIA_PLAYBACK_PAUSE

    DBUS_NAME = "org.gnome.ScreenSaver"
    DBUS_INTERFACE = "org.gnome.ScreenSaver"
    DBUS_PATH = "/org/gnome/ScreenSaver"

    __was_paused = False
    __ignore_next = False
    __interface = None
    __active = False
    __watch = None

    def __on_signal(self, proxy, sender, signal, args):
        if signal == "ActiveChanged":
            # Gnome-Shell fires ActiveChanged even if it doesn't change
            # (aborted transition to lock screen), so handle that
            active = args[0]
            if active == self.__active:
                return
            self.__active = active

            if active:
                self.__was_paused = app.player.paused
                app.player.paused = True
            elif not self.__was_paused and not self.__ignore_next:
                app.player.paused = False

            self.__ignore_next = False

    def plugin_on_unpaused(self):
        # In case pause/unpause happens while the session is inactive
        # (mpris, remote, etc.) don't unpause when it gets active again
        self.__ignore_next = True

    plugin_on_paused = plugin_on_unpaused

    def __remove_interface(self):
        if self.__interface:
            self.__interface.disconnect(self.__sig)
            self.__interface = None

    def __owner_appeared(self, bus, name, owner):
        if not self.__interface:
            iface = Gio.DBusProxy.new_for_bus_sync(
                Gio.BusType.SESSION,
                Gio.DBusProxyFlags.NONE,
                None,
                self.DBUS_NAME,
                self.DBUS_PATH,
                self.DBUS_INTERFACE,
                None,
            )
            self.__sig = iface.connect("g-signal", self.__on_signal)
            self.__active = iface.GetActive()
            self.__interface = iface

    def __owner_vanished(self, bus, owner):
        self.__remove_interface()

    def enabled(self):
        try:
            bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            self.__watch = Gio.bus_watch_name_on_connection(
                bus,
                self.DBUS_NAME,
                Gio.BusNameWatcherFlags.NONE,
                self.__owner_appeared,
                self.__owner_vanished,
            )
        except GLib.Error:
            pass

    def disabled(self):
        if self.__watch:
            Gio.bus_unwatch_name(self.__watch)
        self.__remove_interface()



# Copyright 2010-2011 Christoph Reiter, Steven Robertson
#           2016-2018 Nick Boultbee
#           2018-2019 Peter Strulo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import operator
import os
from functools import reduce

from gi.repository import Gtk, GObject, GLib

import quodlibet
from quodlibet import config
from quodlibet import _

from quodlibet.query import Query
from quodlibet.qltk.cbes import ComboBoxEntrySave
from quodlibet.qltk.ccb import ConfigCheckMenuItem
from quodlibet.qltk.x import SeparatorMenuItem
from quodlibet.qltk import is_accel
from quodlibet.util import limit_songs, DeferredSignal


class SearchBarBox(Gtk.Box):
    """
    A search bar widget for inputting queries.

    signals:
        query-changed - a parsable query string
        focus-out - If the widget gets focused while being focused
            (usually for focusing the songlist)
    """

    __gsignals__ = {
        "query-changed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "focus-out": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    DEFAULT_TIMEOUT = 400

    def __init__(
        self,
        filename=None,
        completion=None,
        accel_group=None,
        timeout=DEFAULT_TIMEOUT,
        validator=Query.validator,
        star=None,
    ):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)

        if filename is None:
            filename = os.path.join(quodlibet.get_user_dir(), "lists", "queries")

        historic_entries_count = config.getint("browsers", "searchbar_historic_entries")

        combo = ComboBoxEntrySave(
            filename,
            count=historic_entries_count,
            validator=validator,
            title=_("Saved Searches"),
            edit_title=_("Edit saved searches…"),
        )

        self.__deferred_changed = DeferredSignal(
            self._filter_changed, timeout=timeout, owner=self
        )

        self.__combo = combo
        entry = combo.get_child()
        self._entry = entry
        if completion:
            entry.set_completion(completion)

        self._star = star
        self._query = None
        self.__sig = combo.connect("text-changed", self.__text_changed)

        entry.connect("clear", self._filter_changed)
        entry.connect("backspace", self.__text_changed)
        entry.connect("populate-popup", self.__menu)
        entry.connect("activate", self._filter_changed)
        entry.connect("activate", self.__save_search)
        entry.connect("focus-out-event", self.__save_search)
        entry.connect("key-press-event", self.__key_pressed)

        entry.set_placeholder_text(_("Search"))
        entry.set_tooltip_text(_("Search your library, using free text or QL queries"))

        combo.enable_clear_button()
        self.pack_start(combo, True, True, 0)

        if accel_group:
            key, mod = Gtk.accelerator_parse("<Primary>L")
            accel_group.connect(key, mod, 0, lambda *x: entry.mnemonic_activate(True))

        for child in self.get_children():
            child.show_all()

    def set_enabled(self, enabled=True):
        self._entry.set_sensitive(enabled)

    def set_text(self, text):
        """Set the text without firing any signals"""

        self.__deferred_changed.abort()
        self._update_query_from(text)

        # deactivate all signals and change the entry text
        self.__inhibit()
        self._entry.set_text(text)
        self.__uninhibit()

    def _update_query_from(self, text):
        # TODO: remove tight coupling to Query
        self._query = Query(text, star=self._star)

    def get_text(self):
        """Get the active text as unicode"""

        return self._entry.get_text()

    def get_query(self, star=None):
        if star and star != self._star:
            self._star = star
            self._update_query_from(self.get_text())
        return self._query

    def changed(self):
        """Triggers a filter-changed signal if the current text
        is a parsable query
        """

        self._filter_changed()

    def __inhibit(self):
        self.__combo.handler_block(self.__sig)

    def __uninhibit(self):
        self.__combo.handler_unblock(self.__sig)

    def __menu(self, entry, menu):
        sep = SeparatorMenuItem()
        sep.show()
        menu.prepend(sep)

        cb = ConfigCheckMenuItem(
            _("Search after _typing"), "settings", "eager_search", populate=True
        )
        cb.set_tooltip_text(_("Show search results after the user stops typing"))
        cb.show()
        menu.prepend(cb)

    def __mnemonic_activate(self, label, group_cycling):
        widget = label.get_mnemonic_widget()
        if widget.is_focus():
            self.emit("focus-out")
            return True
        return None

    def __save_search(self, entry, *args):
        # only save the query on focus-out if eager_search is turned on
        if (
            len(args) > 0
            and args[0]
            and not config.getboolean("settings", "eager_search")
        ):
            return

        text = self.get_text().strip()
        if text and self._query and self._query.is_parsable:
            # Adding the active text to the model triggers a changed signal
            # (get_active is no longer -1), so inhibit
            self.__inhibit()
            self.__combo.prepend_text(text)
            self.__combo.write()
            self.__uninhibit()

    def __key_pressed(self, entry, event):
        if is_accel(event, "<Primary>Return") or is_accel(event, "<Primary>KP_Enter"):
            # Save query on Primary+Return accel, even though the focus is kept
            self.__save_search(entry)
        return False

    def _filter_changed(self, *args):
        self.__deferred_changed.abort()
        text = self.get_text()
        self._update_query_from(text)
        if self._query.is_parsable:
            GLib.idle_add(self.emit, "query-changed", text)

    def __text_changed(self, *args):
        if not self._entry.is_sensitive():
            return
        # the combobox has an active entry selected -> no timeout
        # todo: we need a timeout when the selection changed because
        # of keyboard input (up/down arrows)
        if self.__combo.get_active() != -1:
            self._filter_changed()
            return

        if not config.getboolean("settings", "eager_search"):
            return

        self.__deferred_changed()


class LimitSearchBarBox(SearchBarBox):
    """A version of `SearchBarBox` that allows specifying the limiting and
    weighting of a search."""

    class Limit(Gtk.HBox):
        __gsignals__ = {
            "changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        }

        def __init__(self):
            super().__init__(spacing=3, no_show_all=True)
            label = Gtk.Label(label=_("_Limit:"))
            self.pack_start(label, True, True, 0)

            self.__limit = limit = Gtk.SpinButton()
            self.__limit.connect("value-changed", self.__changed)
            limit.set_numeric(True)
            limit.set_range(0, 9999)
            limit.set_increments(5, 100)
            label.set_mnemonic_widget(limit)
            label.set_use_underline(True)
            self.pack_start(limit, True, True, 0)

            self.__weight = Gtk.CheckButton(label=_("_Weight"), use_underline=True)
            self.__weight.connect("toggled", self.__changed)
            self.pack_start(self.__weight, True, True, 0)

            for child in self.get_children():
                child.show()

        def __changed(self, *args):
            self.emit("changed")

        @property
        def value(self):
            return self.__limit.get_value_as_int()

        @property
        def weighted(self):
            return self.__weight.get_active()

    def __init__(self, show_limit=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__limit = self.Limit()
        self.__limit.set_visible(show_limit)
        self.pack_start(self.__limit, False, True, 0)
        self.__limit.connect("changed", self.__limit_changed)

    def __limit_changed(self, *args):
        self.changed()

    def limit(self, songs):
        if self.__limit.get_visible():
            return limit_songs(songs, self.__limit.value, self.__limit.weighted)
        return songs

    def toggle_limit_widgets(self, button):
        """Toggles the visibility of the limit widget according to `button`"""
        if button.get_active():
            self.__limit.show()
        else:
            self.__limit.hide()
        self.changed()


class MultiSearchBarBox(LimitSearchBarBox):
    """An extension of `LimitSearchBarBox` allowing multiple queries.

    Note: Instances of this class must have their flow_box attribute packed by
    their parents or the multiple queries won't work.
    """

    def __init__(self, *args, show_multi=False, multi_filename=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.multi_filename = (
            os.path.join(quodlibet.get_user_dir(), "lists", "multiqueries")
            if multi_filename is None
            else multi_filename
        )

        self._old_placeholder = self._entry.get_placeholder_text()
        self._old_tooltip = self._entry.get_tooltip_text()

        self._add_button = Gtk.Button.new_from_icon_name(
            "list-add", Gtk.IconSize.BUTTON
        )
        self._add_button.set_no_show_all(True)
        self.pack_start(self._add_button, False, True, 0)
        self._add_button.connect("clicked", self.activated)
        self._entry.connect("activate", self.activated)

        self.flow_box = Gtk.FlowBox(
            no_show_all=True,
            max_children_per_line=99,
            selection_mode=Gtk.SelectionMode.NONE,
        )

        self.toggle_multi_bool(show_multi)

    def activated(self, _):
        if self.flow_box.get_visible():
            text = self.get_text().strip()
            if text == "":  # disallow empty queries
                return
            self.add_query_item(text)
            self.set_text("")
            self._filter_changed()

    def add_query_item(self, text):
        q = QueryItem(text, self._filter_changed)
        q.show()
        self.flow_box.add(q)

    def load(self):
        try:
            with open(self.multi_filename) as f:
                for row in f:
                    self.add_query_item(row.strip())
        except OSError:
            pass

    def save(self):
        if not os.path.isdir(os.path.dirname(self.multi_filename)):
            os.makedirs(os.path.dirname(self.multi_filename))

        with open(self.multi_filename, "w") as f:
            f.writelines(lq.string + "\n" for lq in self.flow_box.get_children())

    def _update_query_from(self, text):
        if self.flow_box.get_visible():
            matches = [lq.query._unpack() for lq in self.flow_box.get_children()]

            self._query = Query(text, star=self._star)
            self._query._match = reduce(operator.and_, matches, self._query._match)
        else:
            super()._update_query_from(text)

    def toggle_multi(self, button):
        """Toggles the multiquery mode according to `button`"""
        self.toggle_multi_bool(button.get_active())

    def toggle_multi_bool(self, multi):
        """Toggles the multiquery mode to the given bool"""
        if multi:
            self._add_button.show()
            self.flow_box.show()

            self._old_placeholder = self._entry.get_placeholder_text()
            self._old_tooltip = self._entry.get_tooltip_text()
            self._entry.set_placeholder_text(_("Add query"))
            self._entry.set_tooltip_text(
                _("Add a QL query or free text to be &ed together")
            )
        else:
            self._add_button.hide()
            self.flow_box.hide()

            self._entry.set_placeholder_text(self._old_placeholder)
            self._entry.set_tooltip_text(self._old_tooltip)
        self.changed()


class QueryItem(Gtk.FlowBoxChild):
    """A FlowBoxChild representing a query"""

    def __init__(self, string, changed_callback):
        super().__init__()

        self.changed_callback = changed_callback
        self.string = string
        self.query = Query(string)

        hbox = Gtk.HBox()
        hbox.pack_start(
            Gtk.Label(string, halign=Gtk.Align.START, margin=6), True, True, 0
        )
        btn = Gtk.Button.new_from_icon_name("window-close", Gtk.IconSize.BUTTON)
        btn.set_relief(Gtk.ReliefStyle.NONE)
        btn.connect("clicked", self.remove)
        hbox.pack_start(btn, False, True, 0)
        frame = Gtk.Frame()
        frame.add(hbox)
        self.add(frame)
        self.show_all()

    def remove(self, _):
        self.destroy()
        self.changed_callback()



# Copyright 2016 0x1777
#        2016-24 Nick Boultbee
#           2017 Didier Villevalois
#           2017 Muges
#           2017 Eyenseo
#           2018 Joschua Gandert
#           2018 Blimmo
#           2018 Olli Helin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from functools import lru_cache
from math import ceil, floor

from gi.repository import Gtk, Gdk, Gst
import cairo

from quodlibet import _, app
from quodlibet import print_w
from quodlibet import util
from quodlibet.plugins import PluginConfig, IntConfProp, ConfProp, BoolConfProp
from quodlibet.plugins.events import EventPlugin
from quodlibet.qltk import Align, add_css
from quodlibet.qltk import Icons
from quodlibet.qltk.seekbutton import TimeLabel
from quodlibet.qltk.tracker import TimeTracker
from quodlibet.qltk import get_fg_highlight_color
from quodlibet.qltk.x import SymbolicIconImage
from quodlibet.util import connect_destroy, print_d
from quodlibet.util.path import uri2gsturi


@lru_cache
def parse_color(value: str | Gdk.RGBA) -> Gdk.RGBA:
    col = Gdk.RGBA()
    if isinstance(value, Gdk.RGBA):
        return value
    if not col.parse(value):
        print_w(f"Invalid Gdk color: {value}")
    print_d(f"Using {col} for {value}")
    return col


class Config:
    _config = PluginConfig(__name__)

    elapsed_color = ConfProp(_config, "elapsed_color", "")
    hover_color = ConfProp(_config, "hover_color", "")
    remaining_color = ConfProp(_config, "remaining_color", "")
    show_current_pos = BoolConfProp(_config, "show_current_pos", False)
    seek_amount = IntConfProp(_config, "seek_amount", 5000)
    max_data_points = IntConfProp(_config, "max_data_points", 3000)
    show_time_labels = BoolConfProp(_config, "show_time_labels", True)
    height_px = IntConfProp(_config, "height_px", 40)


CONFIG = Config()


class WaveformSeekBar(Gtk.Box):
    """A widget containing labels and the seekbar."""

    def __init__(self, player, library):
        super().__init__()

        self._player = player
        self._rms_vals = []
        self._hovering = False

        self._elapsed_label = TimeLabel()
        self._remaining_label = TimeLabel()
        self._waveform_scale = WaveformScale(player)

        self.pack_start(Align(self._elapsed_label, border=6), False, True, 0)
        self.pack_start(self._waveform_scale, True, True, 0)
        self.pack_start(Align(self._remaining_label, border=6), False, True, 0)

        for child in self.get_children():
            child.show_all()
        self.set_time_label_visibility(CONFIG.show_time_labels)

        self._waveform_scale.connect("size-allocate", self._update_redraw_interval)
        self._waveform_scale.connect("motion-notify-event", self._on_mouse_hover)
        self._waveform_scale.connect("leave-notify-event", self._on_mouse_leave)

        self._label_tracker = TimeTracker(player)
        self._label_tracker.connect("tick", self._on_tick_label, player)

        self._redraw_tracker = TimeTracker(player)
        self._redraw_tracker.connect("tick", self._on_tick_waveform, player)

        connect_destroy(player, "seek", self._on_player_seek)
        connect_destroy(player, "song-started", self._on_song_started)
        connect_destroy(player, "song-ended", self._on_song_ended)
        connect_destroy(player, "notify::seekable", self._on_seekable_changed)
        connect_destroy(library, "changed", self._on_song_changed, player)

        self.connect("destroy", self._on_destroy)
        self._update(player)

        if player.info:
            self._create_waveform(player.info, CONFIG.max_data_points)

    def set_time_label_visibility(self, is_visible):
        self._time_labels_visible = is_visible
        if is_visible:
            self._elapsed_label.show()
            self._remaining_label.show()
        else:
            self._elapsed_label.hide()
            self._remaining_label.hide()

    def _create_waveform(self, song, points):
        # Close any existing pipeline to avoid leaks
        self._clean_pipeline()

        if not song.is_file:
            return

        command_template = """
        uridecodebin name=uridec
        ! audioconvert
        ! level name=audiolevel interval={} post-messages=true
        ! fakesink sync=false"""
        interval = int(song("~#length") * 1e9 / points)
        if not interval:
            return
        print_d("Computing data for each %.3f seconds" % (interval / 1e9))

        command = command_template.format(interval)
        pipeline = Gst.parse_launch(command)
        pipeline.get_by_name("uridec").set_property("uri", uri2gsturi(song("~uri")))

        bus = pipeline.get_bus()
        self._bus_id = bus.connect("message", self._on_bus_message, points)
        bus.add_signal_watch()

        pipeline.set_state(Gst.State.PLAYING)

        self._pipeline = pipeline
        self._new_rms_vals = []

    def _on_bus_message(self, bus, message, points):
        force_stop = False
        if message.type == Gst.MessageType.ERROR:
            error, debug = message.parse_error()
            print_d(f"Error received from element {message.src.get_name()}: {error}")
            print_d(f"Debugging information: {debug}")
        elif message.type == Gst.MessageType.ELEMENT:
            structure = message.get_structure()
            if structure.get_name() == "level":
                rms_db = structure.get_value("rms")
                if rms_db:
                    # Calculate average of all channels (usually 2)
                    rms_db_avg = sum(rms_db) / len(rms_db)
                    # Normalize dB value to value between 0 and 1
                    rms = pow(10, (rms_db_avg / 20))
                    self._new_rms_vals.append(rms)
                    if len(self._new_rms_vals) >= points:
                        # The audio might be much longer than we anticipated
                        # and we would get way too many events due to the too
                        # short interval set.
                        force_stop = True
            else:
                print_w(f"Got unexpected message of type {message.type}")

        if message.type == Gst.MessageType.EOS or force_stop:
            self._clean_pipeline()

            # Update the waveform with the new data
            self._rms_vals = self._new_rms_vals
            self._waveform_scale.reset(self._rms_vals)
            self._update_redraw_interval()

            # Clear temporary reference to the waveform data
            del self._new_rms_vals

    def _clean_pipeline(self):
        if hasattr(self, "_pipeline") and self._pipeline:
            self._pipeline.set_state(Gst.State.NULL)
            if self._bus_id:
                bus = self._pipeline.get_bus()
                bus.remove_signal_watch()
                bus.disconnect(self._bus_id)
                self._bus_id = None
            if self._pipeline:
                self._pipeline = None

    def _update_redraw_interval(self, *args):
        if self._player.info and self.is_visible():
            # Must be recomputed when size is changed
            interval = self._waveform_scale.compute_redraw_interval()
            self._redraw_tracker.set_interval(interval)

    def _on_destroy(self, *args):
        self._clean_pipeline()
        self._label_tracker.destroy()
        self._redraw_tracker.destroy()

    def _on_tick_label(self, tracker, player):
        self._update_label(player)

    def _on_tick_waveform(self, tracker, player):
        self._update_waveform(player)

    def _on_seekable_changed(self, player, *args):
        self._update_label(player)

    def _on_player_seek(self, player, song, ms):
        self._update(player)

    def _on_song_changed(self, library, songs, player):
        if not player.info:
            return
        # Check that the currently playing song has changed
        if player.info in songs:
            # Trigger a re-computation of the waveform
            self._create_waveform(player.info, CONFIG.max_data_points)
            self._resize_labels(player.info)
            # Only update the label if some tag value changed
            self._update_label(player)

    def _on_song_started(self, player, song):
        if player.info:
            # Trigger a re-computation of the waveform
            self._create_waveform(player.info, CONFIG.max_data_points)
            self._resize_labels(player.info)

        self._rms_vals.clear()
        self._update(player, True)

    def _on_song_ended(self, player, song, ended):
        self._update(player)

    def _update(self, player, full_redraw=False):
        self._update_label(player)
        self._update_waveform(player, full_redraw)

    def _update_label(self, player):
        if not self._time_labels_visible:
            self.set_sensitive(player.info is not None and player.seekable)
            return

        if player.info:
            if self._hovering:
                # Show the position pointed by the mouse
                position = self._waveform_scale.get_mouse_position()
            else:
                # Show the position of the player (converted in seconds)
                position = player.get_position() / 1000.0
            length = player.info("~#length")
            remaining = length - position

            self._elapsed_label.set_time(position)
            self._remaining_label.set_time(remaining)

            self._elapsed_label.set_disabled(not player.seekable)
            self._remaining_label.set_disabled(not player.seekable)
            self.set_sensitive(player.seekable)
        else:
            self._remaining_label.set_disabled(True)
            self._elapsed_label.set_disabled(True)
            self.set_sensitive(False)

    def _update_waveform(self, player, full_redraw=False):
        if player.info:
            # Position in ms, length in seconds
            position = player.get_position() / 1000.0
            length = player.info("~#length")

            if length != 0:
                self._waveform_scale.set_position(position / length)
            else:
                print_d(f"Length reported as zero for {player.info}")
                self._waveform_scale.set_position(0)

            if position == 0 or full_redraw:
                self._waveform_scale.queue_draw()
            else:
                (x, y, w, h) = self._waveform_scale.compute_redraw_area()
                self._waveform_scale.queue_draw_area(x, y, w, h)
        else:
            self._rms_vals.clear()
            self._waveform_scale.queue_draw()

    def _on_mouse_hover(self, _, event):
        def clamp(a, x, b):
            """Return x if a <= x <= b, else the a or b nearest to x."""
            return min(max(x, a), b)

        width = self._waveform_scale.get_allocation().width
        self._waveform_scale.set_mouse_x_position(clamp(0, event.x, width))

        if self._hovering:
            (x, y, w, h) = self._waveform_scale.compute_hover_redraw_area()
            self._waveform_scale.queue_draw_area(x, y, w, h)
        else:
            self._waveform_scale.queue_draw()

        self._update_label(self._player)
        self._hovering = True

    def _on_mouse_leave(self, _, event):
        self._waveform_scale.set_mouse_x_position(-1)
        self._waveform_scale.queue_draw()

        self._hovering = False
        self._update_label(self._player)

    def _resize_labels(self, song):
        """Resize the labels to make sure there is enough space to display the
        length of the songs.

        This prevents the waveform from changing size when the position changes
        from 9:59 to 10:00 for example."""
        length = util.format_time_display(song("~#length"))

        # Get the width needed to display the length of the song (the text
        # displayed in the labels will always be shorter than that)
        layout = self._remaining_label.get_layout()
        layout.set_text(length, -1)
        width, height = layout.get_pixel_size()

        # Set it as the minimum width of the labels to prevent them from
        # changing width
        self._remaining_label.set_size_request(width, -1)
        self._elapsed_label.set_size_request(width, -1)


class WaveformScale(Gtk.EventBox):
    """The waveform widget."""

    _rms_vals: list[int] = []
    _player = None

    def __init__(self, player):
        super().__init__()
        self._player = player
        self.set_size_request(40, CONFIG.height_px)
        self.position = 0
        self._last_drawn_position = 0
        self.override_background_color(Gtk.StateFlags.NORMAL, Gdk.RGBA(alpha=0))

        self.mouse_position = -1
        self._last_mouse_position = -1
        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK | Gdk.EventMask.SCROLL_MASK)
        self._seeking = False

    @property
    def width(self):
        return self.get_allocation().width

    def reset(self, rms_vals):
        self._rms_vals = rms_vals
        self._seeking = False
        self.queue_draw()

    @classmethod
    def reset_config(cls):
        cls.hover_color.cache_clear()
        cls.remaining_color.cache_clear()
        cls.elapsed_color.cache_clear()

    def compute_redraw_interval(self):
        allocation = self.get_allocation()
        width = allocation.width

        scale_factor = self.get_scale_factor()
        pixel_ratio = float(scale_factor)

        # Compute the coarsest time interval for redraws
        length = self._player.info("~#length")
        if length == 0:
            # The length is 0 for example when playing a stream from
            # Internet radio. If 0 is passed forward as the update interval,
            # UI will freeze as it will try to update continuously.
            # The update interval is usually 1 second so use that instead.
            print_d(
                f"Length is zero for {self._player.info}, "
                "using redraw interval of 1000 ms"
            )
            return 1000
        return length * 1000 / max(width * pixel_ratio, 1)

    def compute_redraw_area(self):
        width = self.width
        last_position_x = self._last_drawn_position * width
        position_x = self.position * width
        return self._compute_redraw_area_between(last_position_x, position_x)

    def compute_hover_redraw_area(self):
        return self._compute_redraw_area_between(
            self._last_mouse_position, self.mouse_position
        )

    def _compute_redraw_area_between(self, x1, x2):
        allocation = self.get_allocation()
        width = allocation.width
        height = allocation.height

        scale_factor = self.get_scale_factor()
        pixel_ratio = float(scale_factor)
        line_width = 1.0 / pixel_ratio

        # Compute the thinnest rectangle to redraw
        x = max(0.0, min(x1, x2) - line_width * 5)
        w = min(width, abs(x2 - x1) + line_width * 10)
        return x, 0.0, w, height

    def draw_waveform(
        self,
        cr,
        width,
        height,
        elapsed_color,
        hover_color,
        remaining_color,
        show_current_pos_config,
    ):
        if width == 0 or height == 0:
            return
        scale_factor = self.get_scale_factor()
        pixel_ratio = float(scale_factor)
        line_width = 1.0 / pixel_ratio

        half_height = self.compute_half_height(height, pixel_ratio)

        value_count = len(self._rms_vals)
        max_value = max(self._rms_vals)
        ratio_width = value_count / (float(width) * pixel_ratio)
        ratio_height = max_value / half_height

        cr.set_line_width(line_width)
        cr.set_line_cap(cairo.LINE_CAP_ROUND)
        cr.set_line_join(cairo.LINE_JOIN_ROUND)

        position_width = self.position * width * pixel_ratio
        mouse_position = self.mouse_position * scale_factor

        hw = line_width / 2.0
        # Avoiding object lookups is slightly faster
        data = self._rms_vals

        # Use the clip rectangles to redraw only what is necessary
        for cx, _cy, cw, _ch in cr.copy_clip_rectangle_list():
            for x in range(
                int(floor(cx * pixel_ratio)), int(ceil((cx + cw) * pixel_ratio)), 1
            ):
                if mouse_position >= 0:
                    if self._seeking:
                        # The user is seeking (holding mousebutton down)
                        fg_color = (
                            elapsed_color if x < mouse_position else remaining_color
                        )
                    elif show_current_pos_config:
                        # Use hover color and elapsed color to display the
                        # current playing position while hovering
                        if x < mouse_position:
                            if x < position_width:
                                fg_color = elapsed_color
                            else:
                                fg_color = hover_color
                        elif x < position_width:
                            fg_color = hover_color
                        else:
                            fg_color = remaining_color
                    else:
                        # The mouse is hovering the seekbar
                        fg_color = (
                            hover_color if x < mouse_position else remaining_color
                        )
                else:
                    fg_color = elapsed_color if x < position_width else remaining_color

                cr.set_source_rgba(*list(fg_color))

                # Basic anti-aliasing / oversampling
                u1 = max(0, int(floor((x - hw) * ratio_width)))
                u2 = min(int(ceil((x + hw) * ratio_width)), len(data))
                val = sum(data[u1:u2]) / (ratio_height * (u2 - u1)) if u1 != u2 else 0.0

                hx = x / pixel_ratio + hw
                cr.move_to(hx, half_height - val)
                cr.line_to(hx, half_height + val)
                cr.stroke()

        self._last_drawn_position = self.position
        self._last_mouse_position = self.mouse_position

    def draw_placeholder(self, cr, width, height, color: Gdk.RGBA):
        if width == 0 or height == 0:
            return
        scale_factor = self.get_scale_factor()
        pixel_ratio = float(scale_factor)
        line_width = 1.0 / pixel_ratio
        position_width = self.position * width * pixel_ratio

        half_height = self.compute_half_height(height, pixel_ratio)
        hw = line_width / 2.0

        cr.set_line_width(line_width)
        cr.set_line_cap(cairo.LINE_CAP_ROUND)
        cr.set_line_join(cairo.LINE_JOIN_ROUND)
        cr.set_source_rgba(*self.elapsed_color(self.get_style_context()))
        cr.move_to(hw, half_height)
        cr.rectangle(hw, half_height - line_width, position_width - hw, line_width * 2)
        cr.fill()

        cr.set_source_rgba(*color)
        cr.move_to(position_width, half_height)
        cr.line_to(width, half_height)
        cr.stroke()

    @staticmethod
    def compute_half_height(height, pixel_ratio):
        # Ensure half_height is in the middle of a pixel (c.f. Cairo's FAQ)
        height_px = int(height * pixel_ratio)
        return (height_px if height_px % 2 else height_px - 1) / pixel_ratio / 2

    def do_draw(self, cr):
        context = self.get_style_context()

        # Get colors
        context.save()
        context.set_state(Gtk.StateFlags.NORMAL)
        bg_color = context.get_background_color(context.get_state())
        context.restore()

        # Paint the background
        cr.set_source_rgba(*list(bg_color))
        cr.paint()

        allocation = self.get_allocation()
        width = allocation.width
        height = allocation.height

        if self._rms_vals:
            self.draw_waveform(
                cr,
                width,
                height,
                self.elapsed_color(context),
                self.hover_color(context),
                self.remaining_color(context),
                CONFIG.show_current_pos,
            )
        else:
            self.draw_placeholder(cr, width, height, self.remaining_color(context))

    @classmethod
    @lru_cache
    def elapsed_color(cls, context: Gtk.StyleContext) -> Gdk.RGBA:
        # Check if the user set a different elapsed color in the config
        return (
            parse_color(text)
            if (text := CONFIG.elapsed_color)
            else get_fg_highlight_color(context)
        )

    @classmethod
    @lru_cache
    def hover_color(cls, context: Gtk.StyleContext) -> Gdk.RGBA:
        if CONFIG.hover_color:
            return parse_color(CONFIG.hover_color)
        opacity = 0.4
        elapsed = cls.elapsed_color(context)
        remaining = cls.remaining_color(context)
        r = (
            opacity * elapsed.alpha * elapsed.red
            + (1 - opacity) * remaining.alpha * remaining.red
        )
        g = (
            opacity * elapsed.alpha * elapsed.green
            + (1 - opacity) * remaining.alpha * remaining.green
        )
        b = (
            opacity * elapsed.alpha * elapsed.blue
            + (1 - opacity) * remaining.alpha * remaining.blue
        )
        a = opacity * elapsed.alpha + (1 - opacity) * remaining.alpha
        return Gdk.RGBA(r, g, b, a)

    @classmethod
    @lru_cache
    def remaining_color(cls, context: Gtk.StyleContext) -> Gdk.RGBA:
        if CONFIG.remaining_color:
            return parse_color(CONFIG.remaining_color)
        default = context.get_color(context.get_state())
        default.alpha = 0.35
        return default

    def do_button_press_event(self, event):
        # Left mouse button
        if event.button == 1 and self._player:
            self._seeking = True
            self.queue_draw()

    def do_button_release_event(self, event):
        # Left mouse button
        if event.button == 1 and self._player:
            ratio = event.x / self.get_allocation().width
            length = self._player.info("~#length")
            self._player.seek(ratio * length * 1000)
            self._seeking = False
            self.queue_draw()
            return True
        return None

    def do_scroll_event(self, event):
        if event.direction == Gdk.ScrollDirection.UP:
            self._player.seek(self._player.get_position() + CONFIG.seek_amount)
            self.queue_draw()
        elif event.direction == Gdk.ScrollDirection.DOWN:
            self._player.seek(self._player.get_position() - CONFIG.seek_amount)
            self.queue_draw()

    def set_position(self, position):
        self.position = position

    def set_mouse_x_position(self, mouse_position):
        """Set the horizontal position of the mouse in pixel"""
        self.mouse_position = mouse_position

    def get_mouse_position(self):
        """Return the position of the song pointed by the mouse in seconds"""
        ratio = self.mouse_position / self.get_allocation().width
        length = self._player.info("~#length")
        return ratio * length


class WaveformSeekBarPlugin(EventPlugin):
    """The plugin class."""

    PLUGIN_ID = "WaveformSeekBar"
    PLUGIN_NAME = _("Waveform Seek Bar")
    PLUGIN_ICON = Icons.GO_JUMP
    PLUGIN_CONFIG_SECTION = __name__
    PLUGIN_DESC = _("∿ A seekbar in the shape of the waveform of the current song.")

    def __init__(self):
        self._bar = None

    def enabled(self):
        self._bar = WaveformSeekBar(app.player, app.librarian)
        self._bar.show()
        app.window.set_seekbar_widget(self._bar)

    def disabled(self):
        app.window.set_seekbar_widget(None)
        self._bar.destroy()
        self._bar = None

    def PluginPreferences(self, parent):
        def colour_changed(c: Gdk.RGBA, config_key: str):
            # This can get parsed back, so we're OK writing it
            string = c.to_string()
            print_d(f"Saving {string} for {config_key}")
            setattr(CONFIG, config_key, string)
            WaveformScale.reset_config()
            # It's nice to refresh the running one
            self._bar._waveform_scale.queue_draw()

        def on_show_pos_toggled(button, *args):
            CONFIG.show_current_pos = button.get_active()

        def seek_amount_changed(spinbox):
            CONFIG.seek_amount = spinbox.get_value_as_int()

        def on_show_time_labels_toggled(button, *args):
            CONFIG.show_time_labels = button.get_active()
            if self._bar is not None:
                self._bar.set_time_label_visibility(CONFIG.show_time_labels)

        def on_height_px_changed(spinbox):
            CONFIG.height_px = spinbox.get_value_as_int()
            if self._bar is not None and self._bar._waveform_scale is not None:
                self._bar._waveform_scale.set_size_request(40, CONFIG.height_px)

        def create_color(label_text, config_item):
            hbox = Gtk.HBox(spacing=6)
            label = Gtk.Label(label=label_text)
            label.set_alignment(0.0, 0.5)
            hbox.pack_start(label, False, True, 0)
            colour = getattr(CONFIG, config_item)
            colour_label = Gtk.Label()
            colour_label.set_alignment(0.0, 0.5)
            colour_label.set_size_request(160, -1)

            def colour_updated(_widget, c: Gdk.RGBA):
                colour_changed(c, config_item)
                colour = c.to_string()
                add_css(
                    colour_label,
                    f"* {{ background-color: {colour}; border:1px solid #666; }}",
                )

            def on_clicked(*args):
                chooser = Gtk.ColorChooserDialog()
                chooser.set_property("use-alpha", True)
                chooser.set_rgba(parse_color(colour))
                chooser.connect("color-activated", colour_updated)
                chooser.connect("response", on_exited)
                chooser.run()

            def on_exited(dialog: Gtk.ColorChooserDialog, code):
                colour_updated(dialog, dialog.get_rgba())
                dialog.destroy()

            button = Gtk.Button()
            button.connect("clicked", on_clicked)
            button.add(SymbolicIconImage(Icons.EDIT, Gtk.IconSize.MENU))
            if colour:
                add_css(
                    colour_label,
                    f"""* {{
                   background-color: {colour};
                   border-radius: 3px;
                   border: 1px solid rgba(128,128,128,0.5);
                }}""",
                )
            hbox.pack_end(button, False, False, 0)
            hbox.pack_end(colour_label, False, True, 0)
            return hbox

        vbox = Gtk.VBox(spacing=9)
        box = create_color(_("Foreground color"), "elapsed_color")
        vbox.pack_start(box, True, True, 0)

        box = create_color(_("Hover color"), "hover_color")
        vbox.pack_start(box, True, True, 0)

        box = create_color(_("Remaining color"), "remaining_color")
        vbox.pack_start(box, True, True, 0)

        sw = Gtk.Switch()
        label = Gtk.Label(_("Show current position whilst hovering"))
        sw.set_active(CONFIG.show_current_pos)
        sw.connect("notify::active", on_show_pos_toggled)
        hbox = Gtk.HBox(spacing=6)
        hbox.pack_start(label, False, True, 0)
        hbox.pack_end(sw, False, True, 0)
        vbox.pack_start(hbox, True, True, 0)

        sw = Gtk.Switch()
        label = Gtk.Label(_("Show time labels"))
        sw.set_active(CONFIG.show_time_labels)
        sw.connect("notify::active", on_show_time_labels_toggled)
        hbox = Gtk.HBox(spacing=6)
        hbox.pack_start(label, False, True, 0)
        hbox.pack_end(sw, False, True, 0)
        vbox.pack_start(hbox, True, True, 0)

        hbox = Gtk.HBox(spacing=6)
        label = Gtk.Label(label=_("Seek amount when scrolling (milliseconds):"))
        hbox.pack_start(label, False, True, 0)
        seek_amount = Gtk.SpinButton(
            adjustment=Gtk.Adjustment(CONFIG.seek_amount, 0, 60000, 1000, 1000, 0)
        )
        seek_amount.set_numeric(True)
        seek_amount.connect("changed", seek_amount_changed)
        hbox.pack_end(seek_amount, False, True, 0)
        vbox.pack_start(hbox, True, True, 0)

        hbox = Gtk.HBox(spacing=6)
        label = Gtk.Label(label=_("Waveform height (pixels):"))
        hbox.pack_start(label, False, True, 0)
        height_px = Gtk.SpinButton(
            adjustment=Gtk.Adjustment(CONFIG.height_px, 40, 400, 10, 10, 0)
        )
        height_px.set_numeric(True)
        height_px.connect("changed", on_height_px_changed)
        hbox.pack_end(height_px, False, True, 0)
        vbox.pack_start(hbox, True, True, 0)

        return vbox



# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import xml.sax
from xml.sax.handler import ContentHandler

from gi.repository import Gtk
from senf import uri2fsn

from quodlibet import _
from quodlibet import app
from quodlibet import util
from quodlibet.qltk import Icons
from quodlibet.qltk.msg import WarningMessage, ErrorMessage
from quodlibet.util.path import normalize_path
from quodlibet.plugins.events import EventPlugin


class RBDBContentHandler(ContentHandler):
    def __init__(self, library):
        ContentHandler.__init__(self)

        self._library = library
        self._current = None
        self._tag = None
        self._changed_songs = []

    def characters(self, content):
        if self._current is not None and self._tag is not None:
            self._current[self._tag] = content

    def startElement(self, name, attrs):
        self._tag = None
        if name == "entry" and attrs.get("type") == "song":
            self._current = {}
        elif name in ("location", "rating", "play-count", "last-played"):
            self._tag = name

    def endElement(self, name):
        self._tag = None
        if name == "entry" and self._current is not None:
            current = self._current
            self._current = None
            if len(current) > 1:
                uri = current.pop("location", "")
                try:
                    filename = uri2fsn(uri)
                except ValueError:
                    return

                self._process_song(normalize_path(filename), current)

    def _process_song(self, path, stats):
        song = self._library.get(path, None)
        if not song:
            return

        has_changed = False

        if "rating" in stats:
            try:
                value = int(stats["rating"]) / 5.0
            except ValueError:
                pass
            else:
                song["~#rating"] = value
                has_changed = True

        if "play-count" in stats:
            try:
                value = int(stats["play-count"])
            except ValueError:
                pass
            else:
                # we could sum them, but that would break on multiple imports
                song["~#playcount"] = value
                has_changed = True

        if "last-played" in stats:
            try:
                value = int(stats["last-played"])
            except ValueError:
                pass
            else:
                if value > song("~#lastplayed", 0):
                    song["~#lastplayed"] = value
                    has_changed = True

        if has_changed:
            self._changed_songs.append(song)

    def finish(self):
        """Call at the end, also returns amount of imported songs"""

        count = len(self._changed_songs)
        self._library.changed(self._changed_songs)
        self._changed_songs = []
        return count


def do_import(parent, library):
    db_path = os.path.expanduser("~/.local/share/rhythmbox/rhythmdb.xml")
    handler = RBDBContentHandler(library)
    try:
        xml.sax.parse(db_path, handler)
    except Exception:
        util.print_exc()
        handler.finish()
        msg = _("Import Failed")
        # FIXME: don't depend on the plugin class here..
        ErrorMessage(parent, RBImport.PLUGIN_NAME, msg).run()
    else:
        count = handler.finish()
        msg = _("Successfully imported ratings and statistics for %d songs") % count
        # FIXME: this is just a warning so it works with older QL
        WarningMessage(parent, RBImport.PLUGIN_NAME, msg).run()


class RBImport(EventPlugin):
    PLUGIN_ID = "rbimport"
    PLUGIN_NAME = _("Rhythmbox Import")
    PLUGIN_DESC = _("Imports ratings and song statistics from Rhythmbox.")
    PLUGIN_ICON = Icons.DOCUMENT_OPEN

    def PluginPreferences(self, *args):
        button = Gtk.Button(label=_("Start Import"))

        def clicked_cb(button):
            do_import(button, app.library)

        button.connect("clicked", clicked_cb)
        return button



# QLScrobbler: an Audioscrobbler client plugin for Quod Libet.
# version 0.11
# (C) 2005-2016 by Joshua Kwan <joshk@triplehelix.org>,
#                  Joe Wreschnig <piman@sacredchao.net>,
#                  Franz Pletyz <fpletz@franz-pletz.org>,
#                  Nicholas J. Michalek <djphazer@gmail.com>,
#                  Steven Robertson <steven@strobe.cc>
#     2012-2024    Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import threading
import time
from hashlib import md5
from urllib.parse import urlencode

from gi.repository import Gtk, GLib

import quodlibet
from quodlibet import _
from quodlibet import const, app, util, qltk
from quodlibet.pattern import Pattern
from quodlibet.query import Query
from quodlibet.plugins.events import EventPlugin
from quodlibet.plugins import PluginConfig
from quodlibet.qltk.entry import ValidatingEntry, UndoEntry
from quodlibet.qltk.msg import Message
from quodlibet.qltk import Icons
from quodlibet.util.dprint import print_d, print_w
from quodlibet.util.picklehelper import pickle_load, pickle_dump, PickleError
from quodlibet.util.urllib import urlopen, UrllibError
from quodlibet.errorreport import errorhook

SERVICES = {
    "Last.fm": "https://post.audioscrobbler.com/",
    "Libre.fm": "https://turtle.libre.fm/",
}

DEFAULT_SERVICE = "Last.fm"
DEFAULT_TITLEPAT = "<title><version| (<version>)>"
DEFAULT_ARTISTPAT = "<artist|<artist>|<composer|<composer>|<performer>>>"

plugin_config = PluginConfig("scrobbler")
defaults = plugin_config.defaults
defaults.set("service", DEFAULT_SERVICE)
defaults.set("titlepat", "")
defaults.set("artistpat", "")
defaults.set("url", "")
defaults.set("username", "")
defaults.set("password", "")
defaults.set("exclude", "")
defaults.set("offline", False)


def config_get_url():
    """Gets the URL for the currently configured service.
    This logic was used often enough to be split out from generic config
    """

    # TODO: share this between the classes better
    service = plugin_config.get("service")
    if service in SERVICES:
        return SERVICES[service]
    return plugin_config.get("url")


def config_get_title_pattern():
    return plugin_config.get("titlepat") or DEFAULT_TITLEPAT


def config_get_artist_pattern():
    return plugin_config.get("artistpat") or DEFAULT_ARTISTPAT


class QLSubmitQueue:
    """Manages the submit queue for scrobbles. Works independently of the
    QLScrobbler plugin being enabled; other plugins may use submit() to queue
    songs for scrobbling.
    """

    CLIENT = "qlb"
    CLIENT_VERSION = const.VERSION
    PROTOCOL_VERSION = "1.2"
    SCROBBLER_CACHE_FILE = os.path.join(quodlibet.get_user_dir(), "scrobbler_cache_v2")

    # These objects are shared across instances, to allow other plugins to
    # queue scrobbles in future versions of QL
    queue: list[dict[str, str]] = []
    changed_event = threading.Event()

    def set_nowplaying(self, song):
        """Send a Now Playing notification."""
        formatted = self._format_song(song)
        if not formatted or self.nowplaying_song == formatted:
            return
        self.nowplaying_song = formatted
        self.nowplaying_sent = False
        self.changed()

    def submit(self, song, timestamp=0):
        """Submit a song. If 'timestamp' is 0, the current time will
        be used."""
        formatted = self._format_song(song)
        if formatted is None:
            return
        if timestamp > 0:
            formatted["i"] = str(timestamp)
        elif timestamp == 0:
            formatted["i"] = str(int(time.time()))
        else:
            # TODO: Forging timestamps for submission from PMPs
            return
        self.queue.append(formatted)
        self.changed()

    def _format_song(self, song):
        """Returns a dict with the keys formatted as required by spec."""
        store = {
            "l": str(song.get("~#length", 0)),
            "n": str(song("~#track")),
            "b": song.comma("album"),
            "m": song("musicbrainz_trackid"),
            "t": self.titlepat.format(song),
            "a": self.artpat.format(song),
        }

        # Spec requires title and artist at minimum
        if not (store.get("a") and store.get("t")):
            return None
        return store

    def __init__(self):
        self.nowplaying_song = None
        self.nowplaying_sent = False
        self.sessionid = None
        self.handshake_event = None
        self.handshake_delay = 1.0
        self.failures = 0
        self.handshake_sent = 0
        self.session_id, self.nowplaying_url, self.submit_url = None, None, None

        self.broken = False

        self.username, self.password, self.base_url = ("", "", "")

        # These need to be set early for _format_song to work
        self.titlepat = Pattern(config_get_title_pattern())
        self.artpat = Pattern(config_get_artist_pattern())

        self._load_queue()

    def _load_queue(self):
        try:
            with open(self.SCROBBLER_CACHE_FILE, "rb") as disk_queue_file:
                disk_queue = pickle_load(disk_queue_file)
            os.unlink(self.SCROBBLER_CACHE_FILE)
            self.queue += disk_queue
        except (OSError, PickleError):
            pass

    @classmethod
    def dump_queue(cls):
        if cls.queue:
            print_d(f"Saving scrobble queue to {cls.SCROBBLER_CACHE_FILE}")
            try:
                with open(cls.SCROBBLER_CACHE_FILE, "wb") as disk_queue_file:
                    pickle_dump(cls.queue, disk_queue_file)
            except (OSError, PickleError) as e:
                print_w(f"Couldn't persist scrobble queue ({e})")

    def _check_config(self):
        user = plugin_config.get("username")
        passw = md5(plugin_config.getbytes("password")).hexdigest()
        url = config_get_url()
        if not user or not passw or not url:
            if self.queue and not self.broken:
                self.quick_dialog(
                    _(
                        "Please visit the Plugins window to set "
                        "QLScrobbler up. Until then, songs will not be "
                        "submitted."
                    ),
                    Gtk.MessageType.INFO,
                )
                self.broken = True
        elif (self.username, self.password, self.base_url) != (user, passw, url):
            self.username, self.password, self.base_url = (user, passw, url)
            self.broken = False
            self.handshake_sent = False
        self.offline = plugin_config.getboolean("offline")
        self.titlepat = Pattern(config_get_title_pattern())
        self.artpat = Pattern(config_get_artist_pattern())

    def changed(self):
        """Signal that settings or queue contents were changed."""
        self._check_config()
        pending = self.queue or (self.nowplaying_song and not self.nowplaying_sent)
        if not self.broken and not self.offline and pending:
            self.changed_event.set()
            return
        self.changed_event.clear()

    def run(self):
        """Submit songs from the queue. Call from a daemon thread."""
        # The spec calls for exponential backoff of failed handshakes, with a
        # minimum of 1m and maximum of 120m delay between attempts.
        self.handshake_sent = False
        self.handshake_event = threading.Event()
        self.handshake_event.set()
        self.handshake_delay = 1

        self.failures = 0

        while True:
            self.changed_event.wait()
            if not self.handshake_sent:
                self.handshake_event.wait()
                if self.send_handshake():
                    self.failures = 0
                    self.handshake_delay = 1
                    self.handshake_sent = True
                else:
                    self.handshake_event.clear()
                    self.handshake_delay = min(self.handshake_delay * 2, 120)
                    GLib.timeout_add(
                        self.handshake_delay * 60 * 1000, self.handshake_event.set
                    )
                    continue
            self.changed_event.wait()
            if self.queue:
                if self.send_submission():
                    self.failures = 0
                else:
                    self.failures += 1
                    if self.failures >= 3:
                        self.handshake_sent = False
            elif self.nowplaying_song and not self.nowplaying_sent:
                self.send_nowplaying()
                self.nowplaying_sent = True
            else:
                # Nothing left to do; wait until something changes
                self.changed_event.clear()

    def send_handshake(self, show_dialog=False):
        # construct url
        stamp = int(time.time())
        auth = md5(
            self.password.encode("utf-8") + str(stamp).encode("utf-8")
        ).hexdigest()
        qp = {
            "hs": "true",
            "p": self.PROTOCOL_VERSION,
            "c": self.CLIENT,
            "v": self.CLIENT_VERSION,
            "u": self.username,
            "a": auth,
            "t": int(stamp),
        }
        qs = urlencode(qp)
        url = f"{self.base_url}/?{qs}"
        print_d(f"Sending handshake to service at {self.base_url}.")

        try:
            resp = urlopen(url)
        except UrllibError:
            if show_dialog:
                self.quick_dialog(
                    _("Could not contact service '%s'.") % util.escape(self.base_url),
                    Gtk.MessageType.ERROR,
                )
            else:
                print_d("Could not contact service. Queueing submissions.")
            return False
        except ValueError:
            self.quick_dialog(
                _("Authentication failed: invalid URL."), Gtk.MessageType.ERROR
            )
            self.broken = True
            return False

        # check response
        lines = resp.read().decode("utf-8", "ignore").rstrip().split("\n")
        status = lines.pop(0)

        if status == "OK":
            self.session_id, self.nowplaying_url, self.submit_url = lines
            self.handshake_sent = True
            print_d(
                f"Session ID: {self.session_id}, "
                f"NP URL: {self.nowplaying_url}, "
                f"Submit URL: {self.submit_url}"
            )
            return True
        print_w(f"Bad handshake status: {status}")
        if status == "BADAUTH":
            self.quick_dialog(
                _("Authentication failed: Invalid username '%s' or bad password.")
                % util.escape(self.username),
                Gtk.MessageType.ERROR,
            )
            self.broken = True
        elif status == "BANNED":
            self.quick_dialog(
                _("Client is banned. Contact the author."), Gtk.MessageType.ERROR
            )
            self.broken = True
        elif status == "BADTIME":
            self.quick_dialog(
                _("Wrong system time. Submissions may fail until it is corrected."),
                Gtk.MessageType.ERROR,
            )
        else:  # "FAILED"
            self.quick_dialog(util.escape(status), Gtk.MessageType.ERROR)
        self.changed()
        return False

    def _check_submit(self, url, data):
        data_str = urlencode(data).encode("ascii")
        try:
            resp = urlopen(url, data_str)
        except OSError:
            print_d("Audioscrobbler server not responding, will try later.")
            return False

        resp_save = resp.read().decode("utf-8", "ignore")
        status = resp_save.rstrip().split("\n")[0]
        print_d(f"Submission status: {status}")
        if status == "OK":
            return True
        if status == "BADSESSION":
            self.handshake_sent = False
            return False
        return False

    def send_submission(self):
        data = {"s": self.session_id}
        to_submit = self.queue[: min(len(self.queue), 50)]
        for idx, song in enumerate(to_submit):
            for key, val in song.items():
                data["%s[%d]" % (key, idx)] = val.encode("utf-8")
            data["o[%d]" % idx] = "P"
            data["r[%d]" % idx] = ""

        song_info = "\n\t".join(f"{s['a']} - {s['t']}" for s in to_submit)
        print_d(f"Submitting song(s): {song_info}")

        if self._check_submit(self.submit_url, data):
            del self.queue[: len(to_submit)]
            return True
        return False

    def send_nowplaying(self):
        data = {"s": self.session_id}
        for key, val in self.nowplaying_song.items():
            data[key] = val.encode("utf-8")
        print_d(
            "Now playing song: "
            f"{self.nowplaying_song['a']} - {self.nowplaying_song['t']}"
        )

        return self._check_submit(self.nowplaying_url, data)

    def quick_dialog_helper(self, dialog_type, msg):
        dialog = Message(dialog_type, app.window, "QLScrobbler", msg)
        dialog.connect("response", lambda dia, resp: dia.destroy())
        dialog.show()

    def quick_dialog(self, msg, dialog_type):
        GLib.idle_add(self.quick_dialog_helper, dialog_type, msg)


class QLScrobbler(EventPlugin):
    PLUGIN_ID = "QLScrobbler"
    PLUGIN_NAME = _("AudioScrobbler Submission")
    PLUGIN_DESC = _(
        "Audioscrobbler client for Last.fm, Libre.fm and other "
        "Audioscrobbler services."
    )
    PLUGIN_ICON = Icons.NETWORK_WORKGROUP

    AUTOSAVE_INTERVAL = 30
    """How often, in seconds, to save the queue (if any) to disk"""

    def __init__(self):
        self.__enabled = False
        self.queue = QLSubmitQueue()
        self._tid = None

        def queue_run():
            try:
                self.queue.run()
            except Exception:
                errorhook()

        queue_thread = threading.Thread(None, queue_run)
        queue_thread.setDaemon(True)
        queue_thread.start()
        print_d(f"Started Scrobbler thread {queue_thread}")

        self.start_time = 0
        self.unpaused_time = 0
        self.elapsed = 0
        self.nowplaying = None

        self.exclude = plugin_config.get("exclude")

    def plugin_on_song_ended(self, song, stopped):
        if song is None or not self.__enabled:
            return
        if self.unpaused_time > 0:
            self.elapsed += time.time() - self.unpaused_time
        # Spec: * don't submit when song length < 00:30
        #       * submit at end of playback (not in the middle, as with v1.1)
        #       * submit if played for >= .5*length or >= 240s
        # we check 'elapsed' rather than 'length' to work around wrong ~#length
        if self.elapsed < 30:
            return
        if self.elapsed < 240 and self.elapsed <= 0.5 * song.get("~#length", 0):
            return
        print_d(f"Checking against filter {self.exclude}")
        if self.exclude and Query(self.exclude).search(song):
            print_d("Not submitting: {}".format(song("~artist~title")))
            return
        self.queue.submit(song, self.start_time)

    def song_excluded(self, song):
        if self.exclude and Query(self.exclude).search(song):
            print_d("{} is excluded by {}".format(song("~artist~title"), self.exclude))
            return True
        return False

    def send_nowplaying(self, song):
        if not self.song_excluded(song):
            self.queue.set_nowplaying(song)

    def plugin_on_song_started(self, song):
        if song is None:
            return
        self.start_time = int(time.time())
        if app.player.paused:
            self.unpaused_time = 0
        else:
            self.unpaused_time = time.time()
        self.elapsed = 0
        if self.__enabled and not app.player.paused:
            self.send_nowplaying(song)
        else:
            self.nowplaying = song

    def plugin_on_paused(self):
        if self.unpaused_time > 0:
            self.elapsed += time.time() - self.unpaused_time
        self.unpaused_time = 0

    def plugin_on_unpaused(self):
        self.unpaused_time = time.time()
        if self.__enabled and self.nowplaying:
            self.send_nowplaying(self.nowplaying)
            self.nowplaying = None

    def enabled(self):
        self.__enabled = True
        print_d("Plugin enabled - accepting new songs.")
        if not self._tid:
            print_d("Starting scrobble autosaver")
            self._tid = GLib.timeout_add_seconds(
                self.AUTOSAVE_INTERVAL, lambda: QLSubmitQueue.dump_queue() or True
            )

    def disabled(self):
        self.__enabled = False
        print_d("Plugin disabled - not accepting any new songs.")
        if self._tid:
            print_d("Stopping scrobble autosaver")
            GLib.source_remove(self._tid)
            self._tid = None
        QLSubmitQueue.dump_queue()

    def PluginPreferences(self, parent):
        def changed(entry, key):
            if entry.get_property("sensitive"):
                plugin_config.set(key, entry.get_text())

        def combo_changed(widget, urlent):
            service = widget.get_active_text()
            plugin_config.set("service", service)
            urlent.set_sensitive(service not in SERVICES)
            urlent.set_text(config_get_url())

        def check_login(*args):
            queue = QLSubmitQueue()
            queue.changed()
            status = queue.send_handshake(show_dialog=True)
            if status:
                queue.quick_dialog(
                    _("Authentication successful."), Gtk.MessageType.INFO
                )

        box = Gtk.VBox(spacing=12)

        # first frame
        table = Gtk.Table(n_rows=5, n_columns=2)
        table.props.expand = False
        table.set_col_spacings(6)
        table.set_row_spacings(6)

        labels = []
        label_names = [_("_Service:"), _("_URL:"), _("User_name:"), _("_Password:")]
        for idx, name in enumerate(label_names):
            label = Gtk.Label(label=name)
            label.set_alignment(0.0, 0.5)
            label.set_use_underline(True)
            table.attach(
                label,
                0,
                1,
                idx,
                idx + 1,
                xoptions=Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK,
            )
            labels.append(label)

        row = 0
        service_combo = Gtk.ComboBoxText()
        table.attach(service_combo, 1, 2, row, row + 1)
        cur_service = plugin_config.get("service")

        # Translators: Other service
        other_label = _("Other…")
        for idx, serv in enumerate(sorted(SERVICES.keys()) + [other_label]):
            service_combo.append_text(serv)
            if cur_service == serv:
                service_combo.set_active(idx)
        if service_combo.get_active() == -1:
            service_combo.set_active(0)
        labels[row].set_mnemonic_widget(service_combo)
        row += 1

        # url
        entry = UndoEntry()
        entry.set_text(plugin_config.get("url"))
        entry.connect("changed", changed, "url")
        service_combo.connect("changed", combo_changed, entry)
        service_combo.emit("changed")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # username
        entry = UndoEntry()
        entry.set_text(plugin_config.get("username"))
        entry.connect("changed", changed, "username")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # password
        entry = UndoEntry()
        entry.set_text(plugin_config.get("password"))
        entry.set_visibility(False)
        entry.connect("changed", changed, "password")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # verify data
        button = qltk.Button(_("_Verify account data"), Icons.DIALOG_INFORMATION)
        button.connect("clicked", check_login)
        table.attach(button, 0, 2, 4, 5)

        box.pack_start(qltk.Frame(_("Account"), child=table), True, True, 0)

        # second frame
        table = Gtk.Table(n_rows=4, n_columns=2)
        table.props.expand = False
        table.set_col_spacings(6)
        table.set_row_spacings(6)

        label_names = [
            _("_Artist pattern:"),
            _("_Title pattern:"),
            _("Exclude _filter:"),
        ]

        labels = []
        for idx, name in enumerate(label_names):
            label = Gtk.Label(label=name)
            label.set_alignment(0.0, 0.5)
            label.set_use_underline(True)
            table.attach(
                label,
                0,
                1,
                idx,
                idx + 1,
                xoptions=Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK,
            )
            labels.append(label)

        row = 0
        # artist pattern
        entry = UndoEntry()
        entry.set_text(plugin_config.get("artistpat"))
        entry.connect("changed", changed, "artistpat")
        table.attach(entry, 1, 2, row, row + 1)
        entry.set_tooltip_text(
            _(
                "The pattern used to format "
                "the artist name for submission. Leave blank for "
                "default."
            )
        )
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # title pattern
        entry = UndoEntry()
        entry.set_text(plugin_config.get("titlepat"))
        entry.connect("changed", changed, "titlepat")
        table.attach(entry, 1, 2, row, row + 1)
        entry.set_tooltip_text(
            _(
                "The pattern used to format "
                "the title for submission. Leave blank for default."
            )
        )
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # exclude filter
        entry = ValidatingEntry(Query.validator)
        entry.set_text(plugin_config.get("exclude"))
        entry.set_tooltip_text(_("Songs matching this filter will not be submitted"))
        entry.connect("changed", changed, "exclude")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # offline mode
        offline = plugin_config.ConfigCheckButton(
            _("_Offline mode (don't submit anything)"), "offline", populate=True
        )
        table.attach(offline, 0, 2, row, row + 1)

        box.pack_start(qltk.Frame(_("Submission"), child=table), True, True, 0)

        return box



# Copyright 2013 Christoph Reiter <reiter.christoph@gmail.com>
#           2024 Nick Boultbee
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
For this plugin to work GNOME Shell needs this file:

/usr/share/gnome-shell/search-providers/io.github.quodlibet.QuodLibet-search-provider.ini
(or in a similar XDG directory)

A copy of this file can be found in ../../../data/
"""

import os
import sys
from pathlib import Path
from collections.abc import Collection

from quodlibet.util.thumbnails import get_thumbnail, get_cache_info

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

from gi.repository import GLib
from gi.repository import Gio

from quodlibet import _, print_d
from quodlibet import app
from quodlibet.util.dbusutils import dbus_unicode_validate
from quodlibet.plugins.events import EventPlugin
from quodlibet.query import Query
from quodlibet.plugins import PluginImportError
from quodlibet.util.path import xdg_get_system_data_dirs
from quodlibet.qltk import Icons

DEFAULT_SEARCH_PROVIDER_DIR = "/usr/share/gnome-shell/search-providers"
THUMBNAIL_SIZE = (256, 256)


def get_gs_provider_files() -> Collection[Path]:
    """Return all installed search provider files for GNOME Shell"""

    ini_files = []
    for d in xdg_get_system_data_dirs():
        path = Path(d) / "gnome-shell" / "search-providers"
        try:
            for entry in os.listdir(path):
                if entry.endswith(".ini"):
                    ini_files.append(path / entry)
        except OSError:
            pass
    return ini_files


def check_ini_installed():
    """Raise if no GNOME Shell ini file for Quod Libet is found"""

    provider_installed = False
    for path in get_gs_provider_files():
        try:
            with open(path, "rb") as handle:
                data = handle.read().decode("utf-8", "replace")
                if SearchProvider.BUS_NAME in data:
                    provider_installed = True
                    break
        except OSError:
            pass

    if not provider_installed:
        path = DEFAULT_SEARCH_PROVIDER_DIR
        msg = (
            _("No GNOME Shell search provider for Quod Libet installed.")
            + " \n"
            + _("Have you copied the ini file to %s (or similar)?") % path
        )
        raise PluginImportError(msg)


class GnomeSearchProvider(EventPlugin):
    PLUGIN_ID = "searchprovider"
    PLUGIN_NAME = _("GNOME Search Provider")
    PLUGIN_DESC = _("Allows GNOME Shell to search the library.")
    PLUGIN_ICON = Icons.SYSTEM_SEARCH

    def enabled(self):
        self.obj = SearchProvider()

    def disabled(self):
        self.obj.remove_from_connection()
        del self.obj

        import gc

        gc.collect()


ENTRY_ICON = ". GThemedIcon audio-mpeg gnome-mime-audio-mpeg audio-x-generic"


def get_song_id(song):
    return str(id(song))


def get_songs_for_ids(library, ids):
    songs = []
    ids = set(ids)
    for song in library:
        song_id = get_song_id(song)
        if song_id in ids:
            songs.append(song)
            ids.discard(song_id)
            if not ids:
                break
    print_d(f"Got {len(songs)} songs matching {ids}")
    return songs


class SearchProvider:
    """
    <!DOCTYPE node PUBLIC
     '-//freedesktop//DTD D-BUS Object Introspection 1.0//EN'
     'http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd'>
    <node name="/io/github/quodlibet/QuodLibet/SearchProvider">
      <interface name="org.freedesktop.DBus.Introspectable">
        <method name="Introspect">
          <arg direction="out" type="s" />
        </method>
      </interface>
      <interface name="org.gnome.Shell.SearchProvider2">
        <method name="GetInitialResultSet">
          <arg direction="in"  type="as" name="terms" />
          <arg direction="out" type="as" name="results" />
        </method>
        <method name="GetSubsearchResultSet">
          <arg direction="in"  type="as" name="previous_results" />
          <arg direction="in"  type="as" name="terms" />
          <arg direction="out" type="as" name="results" />
        </method>
        <method name="GetResultMetas">
          <arg direction="in"  type="as" name="identifiers" />
          <arg direction="out" type="aa{sv}" name="metas" />
        </method>
        <method name="ActivateResult">
          <arg direction="in"  type="s" name="identifier" />
          <arg direction="in"  type="as" name="terms" />
          <arg direction="in"  type="u" name="timestamp" />
        </method>
        <method name="LaunchSearch">
          <arg direction="in"  type="as" name="terms" />
          <arg direction="in"  type="u" name="timestamp" />
        </method>
      </interface>
    </node>
    """

    PATH = "/io/github/quodlibet/QuodLibet/SearchProvider"
    BUS_NAME = "io.github.quodlibet.QuodLibet.SearchProvider"
    IFACE = "org.gnome.Shell.SearchProvider2"

    def __init__(self):
        self._own_id = Gio.bus_own_name(
            Gio.BusType.SESSION,
            self.BUS_NAME,
            Gio.BusNameOwnerFlags.NONE,
            self.on_bus_acquired,
            None,
            self.on_name_lost,
        )
        self._registered_ids = []
        self._method_outargs = {}

    def on_bus_acquired(self, connection, name):
        info = Gio.DBusNodeInfo.new_for_xml(self.__doc__)
        for interface in info.interfaces:
            for method in interface.methods:
                self._method_outargs[method.name] = "({})".format(
                    "".join([arg.signature for arg in method.out_args])
                )

            _id = connection.register_object(
                object_path=self.PATH,
                interface_info=interface,
                method_call_closure=self.on_method_call,
            )
            self._registered_ids.append(_id)

    def on_name_lost(self, connection, name):
        for _id in self._registered_ids:
            connection.unregister_object(_id)

    def remove_from_connection(self):
        if self._own_id is not None:
            Gio.bus_unown_name(self._own_id)
            self._own_id = None

    def on_method_call(
        self,
        connection,
        sender,
        object_path,
        interface_name,
        method_name,
        parameters,
        invocation,
    ):
        args = list(parameters.unpack())
        result = getattr(self, method_name)(*args)
        if not isinstance(result, tuple):
            result = (result,)

        out_args = self._method_outargs[method_name]
        if out_args != "()":
            variant = GLib.Variant(out_args, result)
            invocation.return_value(variant)
        else:
            invocation.return_value(None)

    def Introspect(self):
        return self.__doc__

    def GetInitialResultSet(self, terms):
        print_d(f"Getting initial result set for {terms}")
        if terms:
            query = Query("")
            for term in terms:
                query &= Query(term)
            songs = filter(query.search, app.library)
        else:
            songs = app.library.values()

        return [get_song_id(s) for s in songs]

    def GetSubsearchResultSet(self, previous_results, terms):
        # Eager searching-as-you-type in Gnome makes this useless it seems,
        # so just use the full terms each time.
        return self.GetInitialResultSet(terms)

    def GetResultMetas(self, identifiers):
        print_d(f"Getting result metas for {identifiers}")
        metas = []
        for song in get_songs_for_ids(app.library, identifiers):
            name = song("title")
            description = song.comma("~people")
            song_id = get_song_id(song)
            cover = app.cover_manager.get_cover(song)
            if cover:
                # We need a permanent-ish copy of this for DBus clients
                get_thumbnail(cover.name, THUMBNAIL_SIZE, ignore_temp=False)
                p = get_cache_info(Path(cover.name), THUMBNAIL_SIZE)[0]
                gicon = str(p)
            else:
                gicon = ENTRY_ICON
            meta = {
                "name": GLib.Variant("s", dbus_unicode_validate(name)),
                "id": GLib.Variant("s", song_id),
                "description": GLib.Variant("s", dbus_unicode_validate(description)),
                "gicon": GLib.Variant("s", gicon),
            }
            metas.append(meta)

        return metas

    def ActivateResult(self, identifier, terms, timestamp):
        songs = get_songs_for_ids(app.library, [identifier])
        if not songs:
            return

        if app.player.go_to(songs[0], True):
            app.player.paused = False

    def LaunchSearch(self, terms, timestamp):
        try:
            app.window.browser.filter_text(" ".join(terms))
        except NotImplementedError:
            pass
        else:
            app.present()


# the plugin is useless without the ini file...
check_ini_installed()



# Copyright 2014 Jan Path
#           2014 Christoph Reiter
#           2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import _
from quodlibet import config
from quodlibet.plugins.songshelpers import any_song
from quodlibet.qltk.getstring import GetStringDialog
from quodlibet.qltk import Icons
from quodlibet.qltk.ratingsmenu import ConfirmRateMultipleDialog
from quodlibet.plugins.songsmenu import SongsMenuPlugin


class ExactRating(SongsMenuPlugin):
    PLUGIN_ID = "exact-rating"
    PLUGIN_NAME = _("Set Exact Rating")
    PLUGIN_DESC = _("Allows setting the rating of songs with a number.")
    REQUIRES_ACTION = True
    PLUGIN_ICON = Icons.USER_BOOKMARKS

    plugin_handles = any_song(lambda s: s.can_change())

    def plugin_songs(self, songs):
        value = -1
        while not 0 <= value <= 1:
            input_string = GetStringDialog(
                self.plugin_window,
                self.PLUGIN_NAME,
                _("Please give your desired rating on a scale from 0.0 to 1.0"),
                _("_Apply"),
                Icons.NONE,
            ).run()

            if input_string is None:
                return

            try:
                value = float(input_string)
            except ValueError:
                continue

        count = len(songs)
        if count > 1 and config.getboolean("browsers", "rating_confirm_multiple"):
            confirm_dialog = ConfirmRateMultipleDialog(self.plugin_window, count, value)
            if confirm_dialog.run() != Gtk.ResponseType.YES:
                return

        for song in songs:
            song["~#rating"] = value



# Copyright 2006 Lukas Lalinsky
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import ctypes

from quodlibet.util import load_library

try:
    _libxine, name = load_library(["libxine.so.2", "libxine.so.1"])
except OSError as e:
    raise ImportError(e) from e

if name.endswith("2"):
    _version = 2
else:
    _version = 1


class xine_event_t(ctypes.Structure):
    if _version == 1:
        _fields_ = [
            ("type", ctypes.c_int),
            ("stream", ctypes.c_void_p),
            ("data", ctypes.c_void_p),
            ("data_length", ctypes.c_int),
        ]
    elif _version == 2:
        _fields_ = [
            ("stream", ctypes.c_void_p),
            ("data", ctypes.c_void_p),
            ("data_length", ctypes.c_int),
            ("type", ctypes.c_int),
        ]


class xine_ui_message_data_t(ctypes.Structure):
    _fields_ = [
        ("compatibility_num_buttons", ctypes.c_int),
        ("compatibility_str_len", ctypes.c_int),
        ("compatibility_str", 256 * ctypes.c_char),
        ("type", ctypes.c_int),
        ("explanation", ctypes.c_int),
        ("num_parameters", ctypes.c_int),
        ("parameters", ctypes.c_void_p),
        ("messages", ctypes.c_char),
    ]


# event listener callback type
xine_event_listener_cb_t = ctypes.CFUNCTYPE(
    ctypes.c_void_p, ctypes.c_void_p, ctypes.POINTER(xine_event_t)
)

# event types
XINE_EVENT_UI_PLAYBACK_FINISHED = 1
XINE_EVENT_UI_CHANNELS_CHANGED = 2
XINE_EVENT_UI_SET_TITLE = 3
XINE_EVENT_UI_MESSAGE = 4
XINE_EVENT_FRAME_FORMAT_CHANGE = 5
XINE_EVENT_AUDIO_LEVEL = 6
XINE_EVENT_QUIT = 7
XINE_EVENT_PROGRESS = 8

# stream parameters
XINE_PARAM_SPEED = 1  # see below
XINE_PARAM_AV_OFFSET = 2  # unit: 1/90000 ses
XINE_PARAM_AUDIO_CHANNEL_LOGICAL = 3  # -1 => auto, -2 => off
XINE_PARAM_SPU_CHANNEL = 4
XINE_PARAM_VIDEO_CHANNEL = 5
XINE_PARAM_AUDIO_VOLUME = 6  # 0..100
XINE_PARAM_AUDIO_MUTE = 7  # 1=>mute, 0=>unmute
XINE_PARAM_AUDIO_COMPR_LEVEL = 8  # <100=>off, % compress otherw
XINE_PARAM_AUDIO_AMP_LEVEL = 9  # 0..200, 100=>100% (default)
XINE_PARAM_AUDIO_REPORT_LEVEL = 10  # 1=>send events, 0=> don't
XINE_PARAM_VERBOSITY = 11  # control console output
XINE_PARAM_SPU_OFFSET = 12  # unit: 1/90000 sec
XINE_PARAM_IGNORE_VIDEO = 13  # disable video decoding
XINE_PARAM_IGNORE_AUDIO = 14  # disable audio decoding
XINE_PARAM_IGNORE_SPU = 15  # disable spu decoding
XINE_PARAM_BROADCASTER_PORT = 16  # 0: disable, x: server port
XINE_PARAM_METRONOM_PREBUFFER = 17  # unit: 1/90000 sec
XINE_PARAM_EQ_30HZ = 18  # equalizer gains -100..100
XINE_PARAM_EQ_60HZ = 19  # equalizer gains -100..100
XINE_PARAM_EQ_125HZ = 20  # equalizer gains -100..100
XINE_PARAM_EQ_250HZ = 21  # equalizer gains -100..100
XINE_PARAM_EQ_500HZ = 22  # equalizer gains -100..100
XINE_PARAM_EQ_1000HZ = 23  # equalizer gains -100..100
XINE_PARAM_EQ_2000HZ = 24  # equalizer gains -100..100
XINE_PARAM_EQ_4000HZ = 25  # equalizer gains -100..100
XINE_PARAM_EQ_8000HZ = 26  # equalizer gains -100..100
XINE_PARAM_EQ_16000HZ = 27  # equalizer gains -100..100
XINE_PARAM_AUDIO_CLOSE_DEVICE = 28  # force closing audio device
XINE_PARAM_AUDIO_AMP_MUTE = 29  # 1=>mute, 0=>unmute
XINE_PARAM_FINE_SPEED = 30  # 1.000.000 => normal speed
XINE_PARAM_EARLY_FINISHED_EVENT = 31  # send event when demux finish
XINE_PARAM_GAPLESS_SWITCH = 32  # next stream only gapless swi
XINE_PARAM_DELAY_FINISHED_EVENT = 33  # 1/10sec,0=>disable,-1=>forev

# speeds
XINE_SPEED_PAUSE = 0
XINE_SPEED_SLOW_4 = 1
XINE_SPEED_SLOW_2 = 2
XINE_SPEED_NORMAL = 4
XINE_SPEED_FAST_2 = 8
XINE_SPEED_FAST_4 = 16

# metadata
XINE_META_INFO_TITLE = 0
XINE_META_INFO_COMMENT = 1
XINE_META_INFO_ARTIST = 2
XINE_META_INFO_GENRE = 3
XINE_META_INFO_ALBUM = 4
XINE_META_INFO_YEAR = 5
XINE_META_INFO_VIDEOCODEC = 6
XINE_META_INFO_AUDIOCODEC = 7
XINE_META_INFO_SYSTEMLAYER = 8
XINE_META_INFO_INPUT_PLUGIN = 9

# statuses
XINE_STATUS_IDLE = 0
XINE_STATUS_STOP = 1
XINE_STATUS_PLAY = 2
XINE_STATUS_QUIT = 3

XINE_MSG_NO_ERROR = 0  # (messages to UI)
XINE_MSG_GENERAL_WARNING = 1  # (warning message)
XINE_MSG_UNKNOWN_HOST = 2  # (host name)
XINE_MSG_UNKNOWN_DEVICE = 3  # (device name)
XINE_MSG_NETWORK_UNREACHABLE = 4  # none
XINE_MSG_CONNECTION_REFUSED = 5  # (host name)
XINE_MSG_FILE_NOT_FOUND = 6  # (file name or mrl)
XINE_MSG_READ_ERROR = 7  # (device/file/mrl)
XINE_MSG_LIBRARY_LOAD_ERROR = 8  # (library/decoder)
XINE_MSG_ENCRYPTED_SOURCE = 9  # none
XINE_MSG_SECURITY = 10  # (security message)
XINE_MSG_AUDIO_OUT_UNAVAILABLE = 11  # none
XINE_MSG_PERMISSION_ERROR = 12  # (file name or mrl)
XINE_MSG_FILE_EMPTY = 13  # file is empty
XINE_MSG_AUTHENTICATION_NEEDED = 14  # (mrl, likely http); added in 1.2

# xine_t *xine_new(void)
xine_new = _libxine.xine_new
xine_new.restype = ctypes.c_void_p

# void xine_init(xine_t *self)
xine_init = _libxine.xine_init
xine_init.argtypes = [ctypes.c_void_p]

# void xine_exit(xine_t *self)
xine_exit = _libxine.xine_exit
xine_exit.argtypes = [ctypes.c_void_p]

# void xine_config_load(xine_t *self, const char *cfg_filename)
xine_config_load = _libxine.xine_config_load
xine_config_load.argtypes = [ctypes.c_void_p, ctypes.c_char_p]

# const char *xine_get_homedir(void)
xine_get_homedir = _libxine.xine_get_homedir
xine_get_homedir.restype = ctypes.c_char_p

# xine_audio_port_t *xine_open_audio_driver(xine_t *self, const char *id,
#    void *data)
xine_open_audio_driver = _libxine.xine_open_audio_driver
xine_open_audio_driver.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p]
xine_open_audio_driver.restype = ctypes.c_void_p

# void xine_close_audio_driver(xine_t *self, xine_audio_port_t *driver)
xine_close_audio_driver = _libxine.xine_close_audio_driver
xine_close_audio_driver.argtypes = [ctypes.c_void_p, ctypes.c_void_p]

# xine_stream_t *xine_stream_new(xine_t *self,
#    xine_audio_port_t *ao, xine_video_port_t *vo)
xine_stream_new = _libxine.xine_stream_new
xine_stream_new.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]
xine_stream_new.restype = ctypes.c_void_p

# void xine_close(xine_sxine_event_create_listener_threadtream_t *stream)
xine_close = _libxine.xine_close
xine_close.argtypes = [ctypes.c_void_p]

# int xine_open (xine_stream_t *stream, const char *mrl)
xine_open = _libxine.xine_open
xine_open.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
xine_open.restype = ctypes.c_int

# int xine_play(xine_stream_t *stream, int start_pos, int start_time)
xine_play = _libxine.xine_play
xine_play.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
xine_play.restype = ctypes.c_int

# void xine_stop(xine_stream_t *stream)
xine_stop = _libxine.xine_stop
xine_stop.argtypes = [ctypes.c_void_p]

# void xine_dispose(xine_stream_t *stream)
xine_dispose = _libxine.xine_dispose
xine_dispose.argtypes = [ctypes.c_void_p]

xine_event_new_queue = _libxine.xine_event_new_queue
xine_event_new_queue.argtypes = [ctypes.c_void_p]
xine_event_new_queue.restype = ctypes.c_void_p

# void xine_event_dispose_queue(xine_event_queue_t *queue)
xine_event_dispose_queue = _libxine.xine_event_dispose_queue
xine_event_dispose_queue.argtypes = [ctypes.c_void_p]

# void xine_event_create_listener_thread(xine_event_queue_t *queue,
#    xine_event_listener_cb_t callback,
#    void *user_data)
_xine_event_create_listener_thread = _libxine.xine_event_create_listener_thread
_xine_event_create_listener_thread.argtypes = [
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
]

xine_usec_sleep = _libxine.xine_usec_sleep
xine_usec_sleep.argtypes = [ctypes.c_int]

xine_set_param = _libxine.xine_set_param
xine_set_param.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]

xine_get_param = _libxine.xine_get_param
xine_get_param.argtypes = [ctypes.c_void_p, ctypes.c_int]
xine_get_param.restype = ctypes.c_int

xine_get_meta_info = _libxine.xine_get_meta_info
xine_get_meta_info.argtypes = [ctypes.c_void_p, ctypes.c_int]
xine_get_meta_info.restype = ctypes.c_char_p

xine_get_status = _libxine.xine_get_status
xine_get_status.argtypes = [ctypes.c_void_p]
xine_get_status.restype = ctypes.c_int

_xine_get_pos_length = _libxine.xine_get_pos_length
_xine_get_pos_length.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int),
]

xine_get_version_string = _libxine.xine_get_version_string
xine_get_version_string.restype = ctypes.c_char_p

xine_get_file_extensions = _libxine.xine_get_file_extensions
xine_get_file_extensions.argtypes = [ctypes.c_void_p]
xine_get_file_extensions.restype = ctypes.c_char_p

xine_get_mime_types = _libxine.xine_get_mime_types
xine_get_mime_types.argtypes = [ctypes.c_void_p]
xine_get_mime_types.restype = ctypes.c_char_p

xine_list_input_plugins = _libxine.xine_list_input_plugins
xine_list_input_plugins.argtypes = [ctypes.c_void_p]
xine_list_input_plugins.restype = ctypes.POINTER(ctypes.c_char_p)

xine_check_version = _libxine.xine_check_version
xine_check_version.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
xine_check_version.restype = ctypes.c_int


_callbacks = []


def xine_event_create_listener_thread(queue, callback, user_data):
    cb = xine_event_listener_cb_t(callback)
    _callbacks.append(cb)
    _xine_event_create_listener_thread(queue, cb, user_data)


def xine_get_pos_length(stream):
    _pos_stream = ctypes.c_int()
    _pos_time = ctypes.c_int()
    _length_time = ctypes.c_int()
    result = _xine_get_pos_length(
        stream,
        ctypes.byref(_pos_stream),
        ctypes.byref(_pos_time),
        ctypes.byref(_length_time),
    )
    if result:
        return _pos_stream.value, _pos_time.value, _length_time.value
    return 0, 0, 0



# Copyright 2009-2011 Steven Robertson, Christoph Reiter
#                2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

try:
    from collections import abc
except ImportError:
    import collections as abc  # type: ignore
import subprocess
from enum import Enum
from collections.abc import Iterable
from gi.repository import GLib, Gst

from quodlibet import _, print_d, config
from quodlibet.util.string import decode
from quodlibet.util import is_linux, is_windows
from quodlibet.player import PlayerError


class AudioSinks(Enum):
    """Relevant Gstreamer sink elements"""

    FAKE = "fakesink"

    DIRECTSOUND = "directsoundsink"

    PULSE = "pulsesink"
    """from plugins-good"""

    ALSA = "alsasink"
    """from plugins-base"""

    AUTO = "autoaudiosink"
    """from plugins-good"""

    JACK = "jackaudiosink"
    """from plugins-good"""

    WASAPI = "wasapisink"


def pulse_is_running():
    """Returns whether pulseaudio is running"""

    # If we have a pulsesink we can get the server presence through
    # setting the ready state
    element = Gst.ElementFactory.make(AudioSinks.PULSE.value, None)
    if element is not None:
        element.set_state(Gst.State.READY)
        res = element.get_state(0)[0]
        element.set_state(Gst.State.NULL)
        return res != Gst.StateChangeReturn.FAILURE

    # In case we don't have it call the pulseaudio binary
    try:
        subprocess.check_call(["pulseaudio", "--check"])
    except subprocess.CalledProcessError:
        return False
    except OSError:
        return False
    return True


def jack_is_running() -> bool:
    """:returns: whether Jack is running"""

    element = Gst.ElementFactory.make(AudioSinks.JACK.value, "test sink")
    if element:
        element.set_state(Gst.State.READY)
        res = element.get_state(0)[0]
        element.set_state(Gst.State.NULL)
        return res != Gst.StateChangeReturn.FAILURE
    return False


def link_many(elements: Iterable[Gst.Element]) -> None:
    """Links all elements together
    :raises OSError: if they can't all be linked"""
    last = None
    print_d(
        f"Attempting to link Gstreamer element(s): "
        f"{[type(e).__name__ for e in elements]}"
    )
    for element in elements:
        if last:
            if not Gst.Element.link(last, element):
                raise OSError(f"Failed on element: {type(element).__name__}")
        last = element


def unlink_many(elements):
    last = None
    for element in elements:
        if last:
            if not Gst.Element.unlink(last, element):
                return False
        last = element
    return True


def iter_to_list(func):
    objects = []

    iter_ = func()
    while 1:
        status, value = iter_.next()
        if status == Gst.IteratorResult.OK:
            objects.append(value)
        else:
            break
    return objects


def find_audio_sink() -> tuple[Gst.Element, str]:
    """Get the best audio sink available.

    Returns (element, description) or raises PlayerError.
    """

    def sink_options():
        # People with Jack running probably want it more than any other options
        if config.getboolean("player", "gst_use_jack") and jack_is_running():
            print_d("Using JACK output via Gstreamer")
            return [AudioSinks.JACK]
        if is_windows():
            return [AudioSinks.DIRECTSOUND]
        if is_linux() and pulse_is_running():
            return [AudioSinks.PULSE]
        return [
            AudioSinks.AUTO,
            AudioSinks.PULSE,
            AudioSinks.ALSA,
        ]

    options = sink_options()
    for sink in options:
        element = Gst.ElementFactory.make(sink.value, "player")
        if sink == AudioSinks.JACK and not config.getboolean(
            "player", "gst_jack_auto_connect"
        ):
            # Disable the auto-connection to outputs (e.g. maybe there's scripting)
            element.set_property("connect", "none")
        if element is not None:
            return element, sink.value
    else:
        details = ", ".join(s.value for s in options) if options else "[]"
        raise PlayerError(_("No GStreamer audio sink found. Tried: %s") % details)


def gstreamer_sink(pipeline_desc):
    """Returns a list of unlinked gstreamer elements ending with an audio sink
    and a textual description of the pipeline.

    `pipeline_desc` can be gst-launch syntax for multiple elements
    with or without an audiosink.

    In case of an error, raises PlayerError
    """

    pipe = None
    if pipeline_desc:
        try:
            pipe = [Gst.parse_launch(e) for e in pipeline_desc.split("!")]
        except GLib.GError as e:
            message = e.message
            raise PlayerError(_("Invalid GStreamer output pipeline"), message) from e

    if pipe:
        # In case the last element is linkable with a fakesink
        # it is not an audiosink, so we append the default one
        fake = Gst.ElementFactory.make(AudioSinks.FAKE.value, None)
        try:
            link_many([pipe[-1], fake])
        except OSError:
            pass
        else:
            unlink_many([pipe[-1], fake])
            default_elm, default_desc = find_audio_sink()
            pipe += [default_elm]
            pipeline_desc += " ! " + default_desc
    else:
        elm, pipeline_desc = find_audio_sink()
        pipe = [elm]

    return pipe, pipeline_desc


class TagListWrapper(abc.Mapping):
    def __init__(self, taglist, merge=False):
        self._list = taglist
        self._merge = merge

    def __len__(self):
        return self._list.n_tags()

    def __iter__(self):
        for i in range(len(self)):
            yield self._list.nth_tag_name(i)

    def __getitem__(self, key):
        if not Gst.tag_exists(key):
            raise KeyError

        values = []
        index = 0
        while 1:
            value = self._list.get_value_index(key, index)
            if value is None:
                break
            values.append(value)
            index += 1

        if not values:
            raise KeyError

        if self._merge:
            try:
                return " - ".join(values)
            except TypeError:
                return values[0]

        return values


def parse_gstreamer_taglist(tags):
    """Takes a GStreamer taglist and returns a dict containing only
    numeric and unicode values and str keys."""

    merged = {}
    for key in tags.keys():
        value = tags[key]
        # extended-comment sometimes contains a single vorbiscomment or
        # a list of them ["key=value", "key=value"]
        if key == "extended-comment":
            if not isinstance(value, list):
                value = [value]
            for val in value:
                if not isinstance(val, str):
                    continue
                split = val.split("=", 1)
                sub_key = split[0]
                val = split[-1]
                if sub_key in merged:
                    sub_val = merged[sub_key]
                    if not isinstance(sub_val, str):
                        continue
                    if val not in sub_val.split("\n"):
                        merged[sub_key] += "\n" + val
                else:
                    merged[sub_key] = val
        elif isinstance(value, Gst.DateTime):
            value = value.to_iso8601_string()
            merged[key] = value
        else:
            if isinstance(value, int | float):
                merged[key] = value
                continue

            if isinstance(value, bytes):
                value = decode(value)

            if not isinstance(value, str):
                value = str(value)

            if key in merged:
                merged[key] += "\n" + value
            else:
                merged[key] = value

    return merged


def bin_debug(elements, depth=0, lines=None):
    """Takes a list of gst.Element that are part of a prerolled pipeline, and
    recursively gets the children and all caps between the elements.

    Returns a list of text lines suitable for printing.
    """

    from quodlibet.util.dprint import Colorise

    if lines is None:
        lines = []
    else:
        lines.append(" " * (depth - 1) + "\\")

    for _i, elm in enumerate(elements):
        for pad in iter_to_list(elm.iterate_sink_pads):
            caps = pad.get_current_caps()
            if caps:
                lines.append("{}| {}".format(" " * depth, caps.to_string()))
        name = elm.get_name()
        cls = Colorise.blue(type(elm).__name__.split(".", 1)[-1])
        lines.append("{}|-{} ({})".format(" " * depth, cls, name))

        if isinstance(elm, Gst.Bin):
            children = reversed(iter_to_list(elm.iterate_sorted))
            bin_debug(children, depth + 1, lines)

    return lines
